# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for aqt.tfevent_utils.

Note on test coverage / mutants: get_parsed_tfevents() and get_tfevent_paths()
are currently not covered by unit tests as setting up tfevent test files is
non-trivial. Their functionality has been tested on TFEvent files generated
by experiments, and these functions are unlikely to fail silently.
"""

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax_legacy.utils import tfevent_utils
import numpy as onp


class TfEventUtilsTest(parameterized.TestCase):

  @parameterized.named_parameters(
      dict(
          testcase_name='no_duplicates',
          events=tfevent_utils.EventSeries(
              name='dummy',
              steps=onp.array([0, 1]),
              values=onp.array([0, 1]),
              wall_times=onp.array([0, 1]),
          ),
          exp=tfevent_utils.EventSeries(
              name='dummy',
              steps=onp.array([0, 1]),
              values=onp.array([0, 1]),
              wall_times=None,
          )),
      dict(
          testcase_name='one_duplicate',
          events=tfevent_utils.EventSeries(
              name='test_events',
              steps=onp.array([0, 0]),
              values=onp.array([0, 1]),
              wall_times=onp.array([0, 1]),
          ),
          exp=tfevent_utils.EventSeries(
              name='test_events',
              steps=onp.array([0]),
              values=onp.array([1]),
              wall_times=None,
          )),
      dict(
          testcase_name='one_duplicate_unsorted',
          events=tfevent_utils.EventSeries(
              name='test_events',
              steps=onp.array([0, 0]),
              values=onp.array([1, 0]),
              wall_times=onp.array([1, 0]),
          ),
          exp=tfevent_utils.EventSeries(
              name='test_events',
              steps=onp.array([0]),
              values=onp.array([1]),
              wall_times=None,
          )),
      dict(
          testcase_name='multiple_duplicates',
          events=tfevent_utils.EventSeries(
              name='test_events',
              steps=onp.array([0, 0, 1, 2, 2]),
              values=onp.array([0, 1, 2, 4, 5]),
              wall_times=onp.array([0, 1, 2, 3, 4]),
          ),
          exp=tfevent_utils.EventSeries(
              name='test_events',
              steps=onp.array([0, 1, 2]),
              values=onp.array([1, 2, 5]),
              wall_times=None,
          )),
      dict(
          testcase_name='multiple_duplicates_unsorted',
          events=tfevent_utils.EventSeries(
              name='test_events',
              steps=onp.array([2, 1, 0, 0, 2]),
              values=onp.array([4, 2, 1, 0, 5]),
              wall_times=onp.array([3, 2, 1, 0, 4]),
          ),
          exp=tfevent_utils.EventSeries(
              name='test_events',
              steps=onp.array([0, 1, 2]),
              values=onp.array([1, 2, 5]),
              wall_times=None,
          )),
  )
  def test_sort_and_deduplicate_entries(self, events, exp):
    res = tfevent_utils._sort_and_deduplicate_entries(events)
    self.assertEqual(res.name, exp.name)
    onp.testing.assert_array_equal(res.steps, exp.steps)
    onp.testing.assert_array_equal(res.values, exp.values)
    onp.testing.assert_array_equal(res.wall_times, exp.wall_times)

  def test_sort_and_deduplicate_entries_should_raise_error_when_wall_times_unsorted(
      self):
    events = tfevent_utils.EventSeries(
        name='test_events',
        steps=onp.array([0, 1]),
        values=onp.array([0, 1]),
        wall_times=onp.array([1, 0]))
    with self.assertRaises(ValueError):
      tfevent_utils._sort_and_deduplicate_entries(events)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import functools
from typing import Callable

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax.v2 import aqt_quantizer
from aqt.jax.v2 import calibration
from aqt.jax.v2 import config
from aqt.jax.v2 import stochastic_rounding
from aqt.jax.v2 import tiled_dot_general
from aqt.jax.v2 import utils
import aqt.jax.v2.aqt_dot_general as aqt
from aqt.jax.v2.numerics import int_numerics
from aqt.jax.v2.numerics import no_numerics
from aqt.jax.v2.numerics import numerics
import jax
import jax.numpy as jnp
import numpy as np
import scipy.stats


def _apply_po2_scale(quantizer):
  if quantizer.calibration is None:
    return

  calibration_cls = quantizer.calibration
  # TODO(lew): Remove partial inspection wherever possible.
  # Partial inspection is needed because the current implementation of delayed
  # calibration initialization requires the configuration to be set via
  # functools.partial.
  keywords = {}
  if isinstance(calibration_cls, functools.partial):
    keywords = calibration_cls.keywords
    calibration_cls = calibration_cls.func
  keywords.update(po2_scale=True)
  quantizer.calibration = functools.partial(calibration_cls, **keywords)


def test_jaxpr_dtype(f, dg_raws: list[aqt.DotGeneralRaw], float_dtype):
  """Tests whether dot_generals in f conform to dtypes inside of dg_raws."""

  def jaxpr_to_trityp(jaxpr):
    for eq in jaxpr.eqns:
      # This is where dot_generals hide when one uses custom vjp
      if "fun_jaxpr" in eq.params.keys():
        for rec in jaxpr_to_trityp(eq.params["fun_jaxpr"]):
          yield rec

      if eq.primitive.name == "dot_general":
        [lhs, rhs] = eq.invars
        [out] = eq.outvars
        trityp = (lhs.aval, rhs.aval, out.aval)
        yield trityp

  f_jaxpr = jax.make_jaxpr(f)()
  trityps = [trityp for trityp in jaxpr_to_trityp(f_jaxpr)]
  assert len(trityps) == len(dg_raws)
  for (lhs_sa, rhs_sa, out_sa), dg_raw in zip(trityps, dg_raws):
    # If cfg has None, the type is inherited from the arguments' type.
    def assert_dtype_eq(dtype1, dtype2):
      assert dtype1 == dtype2, f"dtype1 != dtype2: {dtype1=} != {dtype2=}"

    # Using str instead of isinstance to work with adhoc import.
    assert (
        str(type(dg_raw.dg_quantizer))
        == "<class 'aqt.jax.v2.aqt_dot_general.DefaultDotGeneralQuantizer'>"
    ), f"Invalid dg_quantizer type. {str(type(dg_raw.dg_quantizer))=}"
    # assert isinstance(
    #     dg_raw.dg_quantizer, aqt.DefaultDotGeneralQuantizer
    # ), f"Invalid dg_quantizer type. {type(dg_raw.dg_quantizer)=}"

    lhs_dtype = dg_raw.dg_quantizer.lhs.numerics.get_dtype()
    rhs_dtype = dg_raw.dg_quantizer.rhs.numerics.get_dtype()
    assert_dtype_eq(lhs_sa.dtype, lhs_dtype or float_dtype)
    assert_dtype_eq(rhs_sa.dtype, rhs_dtype or float_dtype)
    assert_dtype_eq(out_sa.dtype, dg_raw.dg_accumulator_dtype or float_dtype)


def rand_unif(shape, maxval, seed, dtype=jnp.float32):
  key = jax.random.PRNGKey(seed)
  return jax.random.uniform(
      key=key, shape=shape, minval=-maxval, maxval=maxval, dtype=dtype
  )


# The main test strategy is :
# - Test that FQ is sensible (improve fq_noise_test)
#   - Quantization noise (rounding error) is of proper value
#   - Cliping is correct
#   - Stochastic noise is added correctly.
# - Test that we really use int8/int4/bool in XLA/HLO
# - Test of fq vs dot_general equivalence in po2
# - Test that it all applies to gradients
#
#
# TODO(lew): Tests are a bit incomplete. What we need:
# On top of that we shell test:
#  - Gradient is 1 in the range, 0 outside the abs-max range.


def test_eq(name, a, b):
  assert a.shape == b.shape, (a.shape, b.shape)
  # TODO(lew): use library function.
  mean_err = jnp.mean(jnp.abs(a - b))
  if mean_err != 0.0:
    print("mean_err =", mean_err)
    print(a.shape)
    match a.ndim:
      case 1:
        print(a[:3])
      case 2:
        print(a[:3, :3])
    print("sum =", jnp.sum(a))

    print(b.shape)
    match b.ndim:
      case 1:
        print(b[:3])
      case 2:
        print(b[:3, :3])
    print("sum =", jnp.sum(b))

    print(f"FAIL: {name}")
    assert False


# Test exact equalities.
def _check_result_eq(dgs, *, lhs, rhs, gra):
  good_lr, good_gl, good_gr = None, None, None
  for name, dg, options in dgs:
    # Default settings:
    if "test_gradient" not in options:
      options["test_gradient"] = True
    if "mult" not in options:
      options["mult"] = (1.0, 1.0, 1.0)
    if "check_fwd_lhs_tricky_clip_and_round" not in options:
      options["check_fwd_lhs_tricky_clip_and_round"] = False

    lrf = dg(lhs, rhs)
    lr, backprop = jax.vjp(dg, lhs, rhs)

    gl, gr = backprop(gra) if options["test_gradient"] else (None, None)

    if options["check_fwd_lhs_tricky_clip_and_round"]:
      # Test that all the expected were zero anyway
      where_zero_gradient_expected = lhs < 0
      assert (gl[where_zero_gradient_expected] == 0.0).all()

    good_lr = lr if good_lr is None else good_lr
    good_gl = gl if good_gl is None else good_gl
    good_gr = gr if good_gr is None else good_gr

    test_eq(f"{name}: lr vs lrf", lr, lrf)

    lr_mult, gl_mult, gr_mult = options["mult"]
    test_eq(f"{name}: lr", good_lr, lr / lr_mult)  # forward pass
    if options["test_gradient"]:
      test_eq(f"{name}: gl", good_gl, gl / gl_mult)  # backward pass  # pytype: disable=unsupported-operands
      test_eq(f"{name}: gr", good_gr, gr / gr_mult)  # backward pass  # pytype: disable=unsupported-operands


def fqt_param_dict(s, use_fwd_quant, **kwargs):
  return dict(
      dg=config.fully_quantized(
          use_fwd_quant=use_fwd_quant,
          use_stochastic_rounding=False,
          **kwargs,
      ),
      seed=s,
  )


@utils.flax_slots_kw_only_dataclass
class _TrickyNumerics(numerics.AqtNumerics):
  # Needed because int8 casting would do additional clip and round.
  dtype: None | jnp.dtype = None

  def get_dtype(self):
    return self.dtype

  def get_quant_bound(self) -> jnp.ndarray:
    return jnp.array(1.0)

  def fwd(self, x, context):
    del context
    return jax.lax.round(x, jax.lax.RoundingMethod.TO_NEAREST_EVEN)

  def vjp_fwd(self, x, context):
    res = (x,)
    return self.fwd(x, context), res

  def vjp_bwd(self, res, grad):
    (x,) = res
    ret = grad * (x >= 0)
    return (ret, None)


def _modify_dg(
    readonly_dg: aqt.DotGeneral,
    *,
    lhs_dequant_mode: aqt.DequantMode = aqt.DequantMode.OUTPUT,
    rhs_dequant_mode: aqt.DequantMode = aqt.DequantMode.OUTPUT,
    lhs_calibration_mode: aqt.CalibrationMode = aqt.CalibrationMode.CONTRACTING_AXIS,
    rhs_calibration_mode: aqt.CalibrationMode = aqt.CalibrationMode.CONTRACTING_AXIS,
    use_fwd_quant: None | bool = None,
    disable_rounding: bool = False,
    fwd_lhs_tricky_clip_and_round: bool = False,
    local_aqt: None | aqt.LocalAqt = None,
    use_mid_quant: bool = False,
    clip_gradient: bool = False,
) -> aqt.DotGeneral:
  dg = copy.deepcopy(readonly_dg)
  if fwd_lhs_tricky_clip_and_round:
    # Tricky means that we have zero gradient on x < 0
    dg.fwd.dg_quantizer.lhs.numerics = _TrickyNumerics()
    dg.fwd.dg_accumulator_dtype = None

  def _apply_dequant_mode(c, lhs_dequant_mode, rhs_dequant_mode):
    c.lhs.dequant_mode = lhs_dequant_mode
    c.rhs.dequant_mode = rhs_dequant_mode

  def _apply_calibration_mode(c, lhs_calibration_mode, rhs_calibration_mode):
    c.lhs.calibration_mode = lhs_calibration_mode
    c.rhs.calibration_mode = rhs_calibration_mode

  def _disable_quant_types(c, on_lhs=True, on_rhs=True):
    if on_lhs:
      c.dg_quantizer.lhs.numerics.dtype = None
    if on_rhs:
      c.dg_quantizer.rhs.numerics.dtype = None
    if on_lhs or on_rhs:
      c.dg_accumulator_dtype = None

  disable_lhs_quant = lhs_dequant_mode == aqt.DequantMode.THIS_INPUT
  disable_rhs_quant = rhs_dequant_mode == aqt.DequantMode.THIS_INPUT
  for c in [dg.fwd, dg.dlhs, dg.drhs]:
    # Setting po2_scale is ensuring that fake_quant and full dot_general
    # have the same numerics when scales are power of two (po2).
    # We are passing dims to config so that we can reuse it in fake_quant.
    # Power-of-2 scales allow FQ and AQT to be exactly the same.
    _apply_po2_scale(c.dg_quantizer.lhs)
    _apply_po2_scale(c.dg_quantizer.rhs)

    _apply_dequant_mode(c, lhs_dequant_mode, rhs_dequant_mode)
    _apply_calibration_mode(c, lhs_calibration_mode, rhs_calibration_mode)
    _disable_quant_types(c, disable_lhs_quant, disable_rhs_quant)

  if disable_rounding:
    # If we disable all rounding, we are just testing whether the scales are
    # correct. We don't even need to disable clipping and we are testing
    # that the scales are not too large.
    def disable_quant(c):
      _disable_quant_types(c)
      if isinstance(c.dg_quantizer.lhs.numerics, int_numerics.IntSymmetric):
        c.dg_quantizer.lhs.numerics.round = False
      if isinstance(c.dg_quantizer.rhs.numerics, int_numerics.IntSymmetric):
        c.dg_quantizer.rhs.numerics.round = False

    disable_quant(dg.fwd)
    disable_quant(dg.dlhs)
    disable_quant(dg.drhs)

  if use_fwd_quant is not None:
    if not isinstance(dg.fwd.dg_quantizer.lhs.numerics, no_numerics.NoNumerics):
      dg.drhs.rhs.use_fwd_quant = use_fwd_quant
    if not isinstance(dg.fwd.dg_quantizer.rhs.numerics, no_numerics.NoNumerics):
      dg.dlhs.rhs.use_fwd_quant = use_fwd_quant

  if use_mid_quant:
    config.set_use_mid_quant(
        dg,
        fwd_mid_alpha_both=1.0,
        dlhs_mid_alpha_both=1.0,
        drhs_mid_alpha_both=1.0,
    )

  if local_aqt is not None:
    # Currently we are not supporting local_aqt in fwd pass
    # dg.fwd.local_aqt = local_aqt
    dg.dlhs.local_aqt = local_aqt
    dg.drhs.local_aqt = local_aqt

    # When using abs-max scaling, this should be a no-op.
  if isinstance(dg.fwd.dg_quantizer.lhs.numerics, int_numerics.IntSymmetric):
    dg.fwd.dg_quantizer.lhs.numerics.clip_gradient = clip_gradient
  if isinstance(dg.fwd.dg_quantizer.rhs.numerics, int_numerics.IntSymmetric):
    dg.fwd.dg_quantizer.rhs.numerics.clip_gradient = clip_gradient

  return dg


def _aqt_dg_full_lr_diff(
    lhs_dequant_mode: aqt.DequantMode,
    rhs_dequant_mode: aqt.DequantMode,
    lhs_calibration_mode: aqt.CalibrationMode = aqt.CalibrationMode.CONTRACTING_AXIS,
    rhs_calibration_mode: aqt.CalibrationMode = aqt.CalibrationMode.CONTRACTING_AXIS,
    use_fwd_quant: None | bool = None,
    use_mid_quant: bool = False,
    disable_rounding: bool = False,
    fwd_lhs_tricky_clip_and_round: bool = False,
    local_aqt: None | aqt.LocalAqt = None,
    *,
    readonly_dg: aqt.DotGeneral,
    dims: jax.lax.DotDimensionNumbers,
    clip_gradient: bool = False,
) -> Callable[[jnp.ndarray, jnp.ndarray], jnp.ndarray]:
  dg = _modify_dg(
      readonly_dg,
      lhs_dequant_mode=lhs_dequant_mode,
      rhs_dequant_mode=rhs_dequant_mode,
      lhs_calibration_mode=lhs_calibration_mode,
      rhs_calibration_mode=rhs_calibration_mode,
      use_fwd_quant=use_fwd_quant,
      use_mid_quant=use_mid_quant,
      disable_rounding=disable_rounding,
      fwd_lhs_tricky_clip_and_round=fwd_lhs_tricky_clip_and_round,
      local_aqt=local_aqt,
      clip_gradient=clip_gradient,
  )
  dg = config.set_context(dg, key=jax.random.PRNGKey(4), train_step=None)
  return lambda lhs, rhs: dg(lhs, rhs, dims)


def _aqt_dg_full(
    dequant_mode: aqt.DequantMode,
    calibration_mode: aqt.CalibrationMode = aqt.CalibrationMode.CONTRACTING_AXIS,
    use_fwd_quant: None | bool = None,
    disable_rounding: bool = False,
    fwd_lhs_tricky_clip_and_round: bool = False,
    local_aqt: None | aqt.LocalAqt = None,
    use_mid_quant: bool = False,
    *,
    readonly_dg: aqt.DotGeneral,
    dims: jax.lax.DotDimensionNumbers,
    clip_gradient: bool = False,
) -> Callable[[jnp.ndarray, jnp.ndarray], jnp.ndarray]:
  return _aqt_dg_full_lr_diff(
      lhs_dequant_mode=dequant_mode,
      rhs_dequant_mode=dequant_mode,
      lhs_calibration_mode=calibration_mode,
      rhs_calibration_mode=calibration_mode,
      use_fwd_quant=use_fwd_quant,
      use_mid_quant=use_mid_quant,
      disable_rounding=disable_rounding,
      fwd_lhs_tricky_clip_and_round=fwd_lhs_tricky_clip_and_round,
      local_aqt=local_aqt,
      readonly_dg=readonly_dg,
      dims=dims,
      clip_gradient=clip_gradient,
  )


def _aqt_dg_raw_lr_diff(
    lhs_dequant_mode: aqt.DequantMode,
    rhs_dequant_mode: aqt.DequantMode,
    lhs_calibration_mode: aqt.CalibrationMode = aqt.CalibrationMode.CONTRACTING_AXIS,
    rhs_calibration_mode: aqt.CalibrationMode = aqt.CalibrationMode.CONTRACTING_AXIS,
    *,
    readonly_dg: aqt.DotGeneral,
    dims: jax.lax.DotDimensionNumbers,
) -> Callable[[jnp.ndarray, jnp.ndarray], jnp.ndarray]:
  dg = _modify_dg(
      readonly_dg,
      lhs_dequant_mode=lhs_dequant_mode,
      rhs_dequant_mode=rhs_dequant_mode,
      lhs_calibration_mode=lhs_calibration_mode,
      rhs_calibration_mode=rhs_calibration_mode,
  )
  dg = config.set_context(dg, key=jax.random.PRNGKey(4), train_step=None)
  dg.fwd.dg_quantizer.init_calibration()
  return lambda lhs, rhs: dg.fwd(lhs, rhs, None, None, dims)[0]


def _aqt_dg_raw(
    dequant_mode: aqt.DequantMode,
    calibration_mode: aqt.CalibrationMode = aqt.CalibrationMode.CONTRACTING_AXIS,
    *,
    readonly_dg: aqt.DotGeneral,
    dims: jax.lax.DotDimensionNumbers,
) -> Callable[[jnp.ndarray, jnp.ndarray], jnp.ndarray]:
  return _aqt_dg_raw_lr_diff(
      dequant_mode,
      dequant_mode,
      calibration_mode,
      calibration_mode,
      readonly_dg=readonly_dg,
      dims=dims,
  )


class AqtDotGeneralResearchTest(parameterized.TestCase):

  def test_empty(self):
    t = np.random.normal(size=6).reshape((2, 3))
    np.testing.assert_array_equal(t, t)

  @parameterized.product(
      preserve_zero=[True, False],
      prec=[1, 2, 4, 8],
      v=[0.1, 1000.0],
      seed=list(range(10)),
  )
  def test_fq_noise(self, preserve_zero, prec, v, seed):
    key = jax.random.PRNGKey(seed)
    quantizer = config.quantizer_make(prec)
    if isinstance(quantizer.numerics, int_numerics.IntSymmetric):
      quantizer.numerics.preserve_zero = preserve_zero
      if not preserve_zero:
        quantizer.numerics.dtype = None
    quantizer.calib_shared_axes = (0,)
    sample_size = 10000
    shape = (sample_size,)
    a = jax.random.uniform(key, shape, minval=-v, maxval=v)
    a_fq = aqt_quantizer.make_fake_quant(quantizer)(a)
    bucket_noise = a_fq - a  #  ~ U(-bucket_size/2, bucket_size/2)
    bucket_count = (2**prec - 1) if preserve_zero else (2**prec)
    bucket_size = (v * 2) / bucket_count
    noise = bucket_noise / bucket_size + 0.5  # ~U(0, 1)
    pvalue = scipy.stats.kstest(noise, "uniform").pvalue
    assert pvalue > 0.01

  def test_stochastic_rounding_noise(self):
    repeats = 1000
    num_values = 100000
    shape = (repeats, num_values)

    def assert_clt(noise: jnp.ndarray):
      # Test if the sample mean of the noise conforms to Central Limit Theorem

      # std of a uniform distribution [-0.5, 0.5]
      noise_std = 1 / jnp.sqrt(12.0)
      noise_mean = jnp.mean(noise, axis=1)
      assert jnp.all(noise_mean * jnp.sqrt(num_values) < 4 * noise_std)

    # jax.uniform implementation
    noise_fn = stochastic_rounding.JaxUniform()
    jax_uniform_noise = noise_fn(shape, jax.random.PRNGKey(10))
    assert_clt(jax_uniform_noise)
    # customized more efficient implementation
    noise_fn = stochastic_rounding.RandomCenteredUniform()
    custom_1_noise = noise_fn(shape, jax.random.PRNGKey(11))
    assert_clt(custom_1_noise)

  @parameterized.parameters([
      dict(bits=1),
  ])
  def test_fake_quant(
      self,
      bits=4,
      maxval=10.0,
      shape=(20, 1),
  ):
    quantizer = config.quantizer_make(bits, initialize_calibration=False)
    _apply_po2_scale(quantizer)
    quantizer.init_calibration()
    quantizer.calib_shared_axes = (0,)
    x = jnp.linspace(-maxval, maxval, num=shape[0]).reshape(shape)
    grad = jnp.ones(shape) * 12345.0
    x_fq, backprop = jax.vjp(aqt_quantizer.make_fake_quant(quantizer), x)
    gx_fq = backprop(grad)
    # print(f"x     =\n{x}")
    # print(f"x_fq  =\n{x_fq}")
    # print(f"grad  =\n{grad}")
    # print(f"gx_fq =\n{gx_fq}")
    del x
    del x_fq
    del grad
    del gx_fq

    # TODO(lew): test

  @parameterized.parameters([
      dict(
          # TODO(aqt): Change dlhs_bits to 4bit once
          # https://github.com/google/jax/issues/19682 is fixed.
          dg=config.config_v3(
              fwd_bits=3,
              dlhs_bits=6,
              drhs_bits=5,
              drhs_accumulator_dtype=jnp.int32,  # overwriting the default None
          )
      ),
      dict(dg=config.dot_general_make(None, None)),
      dict(dg=config.dot_general_make(1, 1)),
      dict(dg=config.dot_general_make(1, 2)),
      dict(dg=config.dot_general_make(2, 1)),
      dict(dg=config.dot_general_make(2, 2)),
      dict(dg=config.dot_general_make(8, 8)),
      dict(dg=config.dot_general_make(8, 8), clip_gradient=True),
      dict(
          dg=config.dot_general_make(
              8, 8, dlhs_local_aqt=aqt.LocalAqt(contraction_axis_shard_count=2)
          )
      ),
      dict(
          dg=config.dot_general_make(
              8, 8, drhs_local_aqt=aqt.LocalAqt(contraction_axis_shard_count=2)
          )
      ),
      # That test could fail numerically because bf16
      # can't keep in the product of int8*int8 accurately.
      # It just so happens that this test does not fail but others do.
      # We do this test anyway, to catch jax-compilation-time errors.
      dict(dg=config.dot_general_make(2, 2), dtype=jnp.bfloat16),
      dict(dg=config.dot_general_make(8, 8), dtype=jnp.bfloat16),
      dict(dg=config.dot_general_make(None, 8)),
      dict(dg=config.dot_general_make(8, None)),
      dict(
          dg=fqt_param_dict(s=10, use_fwd_quant=True)["dg"],
          dims=(((0, 2), (1, 0)), ((3, 1), (2, 4))),
          # contraction: 2, 5; batch: 4, 3
          lhs_shape=(2, 3, 5, 4),  # non-contr: 3, 4
          rhs_shape=(5, 2, 4, 6, 3),  # non-contr: 4, 6, 3
          gra_shape=(4, 3, 6),
      ),
      dict(
          dg=fqt_param_dict(
              s=10,
              use_fwd_quant=True,
              dlhs_local_aqt=aqt.LocalAqt(contraction_axis_shard_count=2),
          )["dg"],
          dims=(((0, 2), (1, 0)), ((3, 1), (2, 4))),
          # contraction: 2, 5; batch: 4, 3
          lhs_shape=(2, 3, 5, 4),  # non-contr: 3, 4
          rhs_shape=(5, 2, 4, 6, 3),  # non-contr: 4, 6, 3
          gra_shape=(4, 3, 6),
      ),
      dict(
          dg=config.dot_general_make(2, 2),
          dims=(((0, 2), (1, 0)), ((3, 1), (2, 4))),
          # contraction: 2, 5; batch: 4, 3
          lhs_shape=(2, 3, 5, 4),  # non-contr: 3, 4
          rhs_shape=(5, 2, 4, 6, 3),  # non-contr: 4, 6, 3
          gra_shape=(4, 3, 6),
      ),
      *[fqt_param_dict(s, use_fwd_quant=False) for s in range(10)],
      *[fqt_param_dict(s, use_fwd_quant=True) for s in range(10)],
  ])
  def test_dot_general_calibration_with_contracting_axis(
      self,
      dg: aqt.DotGeneral,
      lhs_maxval=10.0,
      rhs_maxval=20.0,
      gra_maxval=30.0,
      dims=(((1,), (0,)), ((), ())),  # classical matmul
      # lhs_shape=(1, 1),
      # rhs_shape=(1, 2),
      # gra_shape=(1, 2),  # has to be the shape of the output
      # lhs_shape=(1, 2),
      # rhs_shape=(2, 3),
      # gra_shape=(1, 3),  # has to be the shape of the output
      lhs_shape=(10, 20),
      rhs_shape=(20, 30),
      gra_shape=(10, 30),  # has to be the shape of the output
      seed=0,
      dtype=jnp.float32,
      clip_gradient=False,
  ):
    readonly_dg = dg
    del dg

    lhs = rand_unif(lhs_shape, lhs_maxval, seed, dtype)
    rhs = rand_unif(rhs_shape, rhs_maxval, seed + 1, dtype)
    gra = rand_unif(gra_shape, gra_maxval, seed + 2, dtype)

    # Prepare utility functions for test.
    aqt_dg_full = functools.partial(
        _aqt_dg_full,
        readonly_dg=readonly_dg,
        dims=dims,
        clip_gradient=clip_gradient,
    )
    aqt_dg_raw = functools.partial(
        _aqt_dg_raw, readonly_dg=readonly_dg, dims=dims
    )
    modify_dg = functools.partial(_modify_dg, readonly_dg=readonly_dg)
    check = functools.partial(_check_result_eq, lhs=lhs, rhs=rhs, gra=gra)

    # Tests for dot_general.
    test_jaxpr_dtype(
        lambda: aqt_dg_full(aqt.DequantMode.OUTPUT)(lhs, rhs),
        [modify_dg().fwd],
        lhs.dtype,
    )
    test_jaxpr_dtype(
        lambda: jax.vjp(aqt_dg_full(aqt.DequantMode.OUTPUT), lhs, rhs),
        [modify_dg().fwd],
        lhs.dtype,
    )
    _, backprop = jax.vjp(aqt_dg_full(aqt.DequantMode.OUTPUT), lhs, rhs)
    test_jaxpr_dtype(
        lambda: backprop(gra),
        [modify_dg().dlhs, modify_dg().drhs],
        gra.dtype,
    )

    check([
        ("default    ", aqt_dg_full(aqt.DequantMode.OUTPUT), dict()),
        ("FQ         ", aqt_dg_full(aqt.DequantMode.THIS_INPUT), dict()),
        (
            "raw fwd    ",
            aqt_dg_raw(aqt.DequantMode.OUTPUT),
            dict(test_gradient=False),
        ),
        (
            "raw fwd FQ ",
            aqt_dg_raw(aqt.DequantMode.THIS_INPUT),
            dict(test_gradient=False),
        ),
    ])

    check([
        (
            "midQ FQ         ",
            aqt_dg_full(
                aqt.DequantMode.THIS_INPUT,
                use_mid_quant=True,
                use_fwd_quant=False,
            ),
            dict(),
        ),
        (
            "midQ       ",
            aqt_dg_full(
                aqt.DequantMode.OUTPUT,
                use_mid_quant=True,
                use_fwd_quant=False,
            ),
            dict(),
        ),
    ])

    check([
        (
            "fwd_quant=F",
            aqt_dg_full(
                aqt.DequantMode.OUTPUT,
                use_fwd_quant=False,
                disable_rounding=True,
            ),
            dict(),
        ),
        (
            "fwd_quant=T",
            aqt_dg_full(
                aqt.DequantMode.OUTPUT,
                use_fwd_quant=True,
                disable_rounding=True,
            ),
            dict(),
        ),
    ])

    check([
        (
            "default    ",
            aqt_dg_full(
                aqt.DequantMode.OUTPUT,
                local_aqt=aqt.LocalAqt(contraction_axis_shard_count=2),
            ),
            dict(),
        ),
        (
            "default    ",
            aqt_dg_full(
                aqt.DequantMode.THIS_INPUT,
                local_aqt=aqt.LocalAqt(contraction_axis_shard_count=2),
            ),
            dict(),
        ),
    ])

    if isinstance(
        readonly_dg.fwd.dg_quantizer.lhs.numerics,
        int_numerics.IntSymmetric,
    ):
      check([
          (
              "check_fwd_lhs_tricky_clip_and_round",
              aqt_dg_full(
                  aqt.DequantMode.OUTPUT, fwd_lhs_tricky_clip_and_round=True
              ),
              dict(check_fwd_lhs_tricky_clip_and_round=True),
          ),
      ])

    def unquant_aqt_dg(lhs, rhs):
      dg = config.default_unquantized_config()
      return dg(lhs, rhs, dims)

    def lax_dg(lhs, rhs):
      return jax.lax.dot_general(lhs, rhs, dims)

    check([
        ("unquantized default:", unquant_aqt_dg, dict()),
        ("lax.dot_general:", lax_dg, dict()),
    ])

  @parameterized.parameters([
      dict(
          dg=lambda: config.config_v3(
              fwd_bits=3,
              dlhs_bits=4,
              drhs_bits=5,
              drhs_accumulator_dtype=jnp.int32,  # overwriting the default None
          )
      ),
      dict(dg=config.dot_general_make(None, None)),
      dict(dg=config.dot_general_make(1, 1)),
      dict(dg=config.dot_general_make(1, 2)),
      dict(dg=config.dot_general_make(2, 1)),
      dict(dg=config.dot_general_make(2, 2)),
      dict(dg=config.dot_general_make(8, 8)),
      dict(dg=config.dot_general_make(8, 8), clip_gradient=True),
      dict(
          dg=config.dot_general_make(
              8, 8, dlhs_local_aqt=aqt.LocalAqt(contraction_axis_shard_count=2)
          )
      ),
      dict(
          dg=config.dot_general_make(
              8, 8, drhs_local_aqt=aqt.LocalAqt(contraction_axis_shard_count=2)
          )
      ),
      # That test could fail numerically because bf16
      # can't keep in the product of int8*int8 accurately.
      # It just so happens that this test does not fail but others do.
      # We do this test anyway, to catch jax-compilation-time errors.
      dict(dg=config.dot_general_make(2, 2), dtype=jnp.bfloat16),
      dict(dg=config.dot_general_make(8, 8), dtype=jnp.bfloat16),
      dict(dg=config.dot_general_make(None, 8)),
      dict(dg=config.dot_general_make(8, None)),
      dict(
          dg=config.dot_general_make(2, 2),
          dims=(((0, 2), (1, 0)), ((3, 1), (2, 4))),
          # contraction: 2, 5; batch: 4, 3
          lhs_shape=(2, 3, 5, 4),  # non-contr: 3, 4
          rhs_shape=(5, 2, 4, 6, 3),  # non-contr: 4, 6, 3
          gra_shape=(4, 3, 6),
      ),
      *[fqt_param_dict(s, use_fwd_quant=False) for s in range(10)],
  ])
  def test_dot_general_calibration_with_remaining_axis(
      self,
      dg: config.DotGeneral | Callable[[], config.DotGeneral],
      lhs_maxval=10.0,
      rhs_maxval=20.0,
      gra_maxval=30.0,
      dims=(((1,), (0,)), ((), ())),  # classical matmul
      lhs_shape=(10, 20),
      rhs_shape=(20, 30),
      gra_shape=(10, 30),  # has to be the shape of the output
      seed=0,
      dtype=jnp.float32,
      clip_gradient=False,
  ):
    # Deferred evaluation of config function calls. 4-bit config initialization
    # triggers jax.local_devices(), which shouldn't be called before
    # absl.app.run() in some environments.
    if not isinstance(dg, config.DotGeneral):
      dg = dg()
    # Set use_fwd_quant to False.
    dg.drhs.rhs.use_fwd_quant = False
    dg.dlhs.rhs.use_fwd_quant = False
    readonly_dg = dg
    del dg

    lhs = rand_unif(lhs_shape, lhs_maxval, seed, dtype)
    rhs = rand_unif(rhs_shape, rhs_maxval, seed + 1, dtype)
    gra = rand_unif(gra_shape, gra_maxval, seed + 2, dtype)

    # Prepare utility functions for test.
    aqt_dg_full = functools.partial(
        _aqt_dg_full,
        readonly_dg=readonly_dg,
        dims=dims,
        clip_gradient=clip_gradient,
    )
    aqt_dg_full_lr_diff = functools.partial(
        _aqt_dg_full_lr_diff,
        readonly_dg=readonly_dg,
        dims=dims,
        clip_gradient=clip_gradient,
    )
    aqt_dg_raw = functools.partial(
        _aqt_dg_raw, readonly_dg=readonly_dg, dims=dims
    )
    aqt_dg_raw_lr_diff = functools.partial(
        _aqt_dg_raw_lr_diff, readonly_dg=readonly_dg, dims=dims
    )
    check = functools.partial(_check_result_eq, lhs=lhs, rhs=rhs, gra=gra)

    # test dot_general
    check([
        (
            "RA L       ",
            aqt_dg_full_lr_diff(
                aqt.DequantMode.OTHER_INPUT,
                aqt.DequantMode.THIS_INPUT,
                aqt.CalibrationMode.REMAINING_AXIS,
                aqt.CalibrationMode.REMAINING_AXIS,
            ),
            dict(),
        ),
        (
            "RA R       ",
            aqt_dg_full_lr_diff(
                aqt.DequantMode.THIS_INPUT,
                aqt.DequantMode.OTHER_INPUT,
                aqt.CalibrationMode.REMAINING_AXIS,
                aqt.CalibrationMode.REMAINING_AXIS,
            ),
            dict(),
        ),
        (
            "RA fake    ",
            aqt_dg_full(
                aqt.DequantMode.THIS_INPUT,
                aqt.CalibrationMode.REMAINING_AXIS,
            ),
            dict(),
        ),
        (
            "RA L fwd   ",
            aqt_dg_raw_lr_diff(
                aqt.DequantMode.OTHER_INPUT,
                aqt.DequantMode.THIS_INPUT,
                aqt.CalibrationMode.REMAINING_AXIS,
                aqt.CalibrationMode.REMAINING_AXIS,
            ),
            dict(test_gradient=False),
        ),
        (
            "RA R fwd  ",
            aqt_dg_raw_lr_diff(
                aqt.DequantMode.THIS_INPUT,
                aqt.DequantMode.OTHER_INPUT,
                aqt.CalibrationMode.REMAINING_AXIS,
                aqt.CalibrationMode.REMAINING_AXIS,
            ),
            dict(test_gradient=False),
        ),
        (
            "RA fake fwd",
            aqt_dg_raw(
                aqt.DequantMode.THIS_INPUT,
                aqt.CalibrationMode.REMAINING_AXIS,
            ),
            dict(test_gradient=False),
        ),
    ])

    check([
        (
            "RA fake    ",
            aqt_dg_full(
                aqt.DequantMode.THIS_INPUT,
                aqt.CalibrationMode.REMAINING_AXIS,
                local_aqt=aqt.LocalAqt(contraction_axis_shard_count=2),
            ),
            dict(),
        ),
        (
            "RA L      ",
            aqt_dg_full_lr_diff(
                aqt.DequantMode.OTHER_INPUT,
                aqt.DequantMode.THIS_INPUT,
                aqt.CalibrationMode.REMAINING_AXIS,
                aqt.CalibrationMode.REMAINING_AXIS,
                local_aqt=aqt.LocalAqt(contraction_axis_shard_count=2),
            ),
            dict(),
        ),
        (
            "RA R       ",
            aqt_dg_full_lr_diff(
                aqt.DequantMode.THIS_INPUT,
                aqt.DequantMode.OTHER_INPUT,
                aqt.CalibrationMode.REMAINING_AXIS,
                aqt.CalibrationMode.REMAINING_AXIS,
                local_aqt=aqt.LocalAqt(contraction_axis_shard_count=2),
            ),
            dict(),
        ),
    ])

  def test_dot_general_calibrate_dequant_mode_mismatch(self):
    dg = config.dot_general_make(8, 8, use_fwd_quant=False)
    dims = (((1,), (0,)), ((), ()))
    lhs = rand_unif((10, 20), 10.0, 0, jnp.float32)
    rhs = rand_unif((20, 30), 20.0, 1, jnp.float32)

    # 1. Raise error when OTHER_INPUT + CONTRACTING_AXIS
    with self.assertRaisesRegex(
        AssertionError,
        ".*Unsupported calibration mode.*dequant mode combination.*",
    ):
      _aqt_dg_full_lr_diff(
          aqt.DequantMode.THIS_INPUT,
          aqt.DequantMode.OTHER_INPUT,
          aqt.CalibrationMode.CONTRACTING_AXIS,
          aqt.CalibrationMode.CONTRACTING_AXIS,
          readonly_dg=copy.deepcopy(dg),
          dims=dims,
      )(lhs, rhs)

    # 2. Raise error when OUTPUT + REMAINING_AXIS
    with self.assertRaisesRegex(
        AssertionError,
        ".*Unsupported calibration mode.*dequant mode combination.*",
    ):
      _aqt_dg_full_lr_diff(
          aqt.DequantMode.THIS_INPUT,
          aqt.DequantMode.OUTPUT,
          aqt.CalibrationMode.CONTRACTING_AXIS,
          aqt.CalibrationMode.REMAINING_AXIS,
          readonly_dg=copy.deepcopy(dg),
          dims=dims,
      )(lhs, rhs)

  def test_dot_general_prevent_fwd_quant_with_remaining_axis(self):
    """If calibration axis is remaining_axis, use_fwd_quant should be False."""
    dg = config.dot_general_make(8, 8, use_fwd_quant=True)
    dims = (((1,), (0,)), ((), ()))
    lhs = rand_unif((10, 20), 10.0, 0, jnp.float32)
    rhs = rand_unif((20, 30), 20.0, 1, jnp.float32)

    with self.assertRaisesRegex(
        AssertionError,
        ".*use_fwd_quant should be set to False.*",
    ):
      _aqt_dg_full_lr_diff(
          aqt.DequantMode.THIS_INPUT,
          aqt.DequantMode.OTHER_INPUT,
          aqt.CalibrationMode.CONTRACTING_AXIS,
          aqt.CalibrationMode.REMAINING_AXIS,
          readonly_dg=copy.deepcopy(dg),
          dims=dims,
      )(lhs, rhs)

  @parameterized.parameters([
      dict(dg=config.dot_general_make(8, 8)),
      dict(
          dg=config.dot_general_make(
              8, 8, dlhs_local_aqt=aqt.LocalAqt(contraction_axis_shard_count=2)
          )
      ),
      dict(
          dg=config.dot_general_make(
              8, 8, drhs_local_aqt=aqt.LocalAqt(contraction_axis_shard_count=2)
          )
      ),
  ])
  def test_dot_general_equality_between_different_calibration_axes(
      self,
      dg: config.DotGeneral,
  ):
    """Check equality between different calibration axes."""
    dims = (((1,), (0,)), ((), ()))

    # Set use_fwd_quant to False.
    dg.drhs.rhs.use_fwd_quant = False
    dg.dlhs.rhs.use_fwd_quant = False
    readonly_dg = dg
    del dg

    # Set the two arguments as powers of 2, to prevent it from having the
    # quantization loss.
    lhs = 2 ** jnp.floor(rand_unif((10, 20), 3.0, 1, jnp.float32) + 3.0)
    rhs = 2 ** jnp.floor(rand_unif((20, 30), 3.0, 2, jnp.float32) + 3.0)
    gra = 2 ** jnp.floor(rand_unif((10, 30), 3.0, 3, jnp.float32) + 3.0)

    # Prepare utility functions for test.
    aqt_dg_full = functools.partial(
        _aqt_dg_full, readonly_dg=readonly_dg, dims=dims
    )
    aqt_dg_full_lr_diff = functools.partial(
        _aqt_dg_full_lr_diff, readonly_dg=readonly_dg, dims=dims
    )
    check = functools.partial(_check_result_eq, lhs=lhs, rhs=rhs, gra=gra)

    # test dot_general
    check([
        (
            "CA         ",
            aqt_dg_full(
                aqt.DequantMode.OUTPUT,
                use_fwd_quant=False,
                disable_rounding=True,
            ),
            dict(),
        ),
        (
            "CA fake    ",
            aqt_dg_full(
                aqt.DequantMode.THIS_INPUT,
                use_fwd_quant=False,
                disable_rounding=True,
            ),
            dict(),
        ),
        (
            "RA L       ",
            aqt_dg_full_lr_diff(
                aqt.DequantMode.OTHER_INPUT,
                aqt.DequantMode.THIS_INPUT,
                aqt.CalibrationMode.REMAINING_AXIS,
                aqt.CalibrationMode.REMAINING_AXIS,
            ),
            dict(),
        ),
        (
            "RA R       ",
            aqt_dg_full_lr_diff(
                aqt.DequantMode.THIS_INPUT,
                aqt.DequantMode.OTHER_INPUT,
                aqt.CalibrationMode.REMAINING_AXIS,
                aqt.CalibrationMode.REMAINING_AXIS,
            ),
            dict(),
        ),
        (
            "RA fake      ",
            aqt_dg_full(
                aqt.DequantMode.THIS_INPUT,
                aqt.CalibrationMode.REMAINING_AXIS,
            ),
            dict(),
        ),
    ])

  def test_dynamic_context(self):
    @jax.jit
    def f(lhs, rhs):
      dg = config.dot_general_make()
      dg = config.set_context(dg, key=jax.random.PRNGKey(4), train_step=None)
      return dg(lhs, rhs, (((0,), (0,)), ((), ())))

    lhs, rhs = jnp.array([3.0, 4.0]), jnp.array([4.0, 5.0])
    jax.value_and_grad(f)(lhs, rhs)

  def test_hardware_int8(self, seed=0):
    dg_raw = config.dot_general_raw_make(8, 8)

    def dg(lhs, rhs):
      ret, _ = dg_raw(
          lhs,
          rhs,
          None,
          None,
          (((1,), (0,)), ((), ())),
      )
      return ret

    lhs = rand_unif((10, 20), 1.0, seed)
    rhs = rand_unif((20, 30), 1.0, seed + 1)
    test_jaxpr_dtype(lambda: dg(lhs, rhs), [dg_raw], lhs.dtype)
    assert dg_raw.dg_quantizer.lhs.numerics.get_dtype() == jnp.int8
    assert dg_raw.dg_quantizer.rhs.numerics.get_dtype() == jnp.int8

  @parameterized.parameters([
      dict(
          shard_count=2,
          lhs=[1270.0, 10.0, 1270000.0, 10000.0],
          expected_product=1281280.0,
      ),
      dict(
          shard_count=1,
          lhs=[1270.0, 10.0, 1270000.0, 10000.0],
          expected_product=1280000.0,
      ),
  ])
  def test_local_aqt(self, shard_count, lhs, expected_product):
    # create a config that quantizes both forward and backward passes to int8
    # set the number of shards (local aqt) to 2
    dg = config.fully_quantized(
        fwd_bits=8,
        bwd_bits=8,
        use_stochastic_rounding=False,
        drhs_local_aqt=aqt.LocalAqt(contraction_axis_shard_count=shard_count),
    )
    dg.fwd.dg_quantizer.lhs.numerics.preserve_max_val = True
    dg.fwd.dg_quantizer.rhs.numerics.preserve_max_val = True
    dg.drhs.dg_quantizer.lhs.numerics.preserve_max_val = True
    dg.drhs.dg_quantizer.rhs.numerics.preserve_max_val = True
    dg_f = lambda lhs, rhs: dg(
        lhs,
        rhs,
        dimension_numbers=(((), ()), ((), ())),
    )
    lhs = jnp.array(lhs)
    rhs = jnp.array([1.0])
    output, bprop = jax.vjp(dg_f, lhs, rhs)
    _, drhs = bprop(jnp.ones_like(output))
    assert drhs == expected_product

  def test_per_tensor(self):
    # TODO(lew): bits=8 started failing in VLP colab due x/x != 1.0 sometimes
    bits = 4
    my_numerics = int_numerics.IntSymmetric(
        bits=bits,
        preserve_zero=True,
        preserve_max_val=False,
        clip=True,
        clip_gradient=False,
        round=True,
        noise_fn=None,
        dtype=jnp.int8,
    )
    quantizer = aqt_quantizer.Quantizer(
        numerics=my_numerics,
        calib_shared_axes="per_tensor",
        scale_stop_grad=True,
        calibration=calibration.AbsMaxCalibration,
        context=utils.Context(key=None, train_step=None),
    )
    # TODO(lew): Perhaps post_init call could work?
    quantizer.init_calibration()

    x = jnp.array([
        [1, 2, 3, 4],
        [5, 6, 7, 2 ** (bits - 1) - 0.5],
        [3, 5, 1, -1],
    ])
    qx, _ = quantizer.quant(x, calibration_axes=None)
    self.assertEqual(qx.scale, [jnp.array([[1.0]])])

  def test_per_subchannel(self):
    # TODO(lew): bits=8 started failing in VLP colab due x/x != 1.0 sometimes
    bits = 4

    # NOTE: The scale dtype must be set to a float dtype when quantizing an
    # integer input, as jax does not support taking the inverse of an integer.
    quantizer = aqt_quantizer.quantizer_make(bits, scale_dtype=jnp.float32)
    x = jnp.arange(0, 64).reshape((4, 4, 4))

    tiling_state = tiled_dot_general.generate_tiling_state(
        x,
        [tiled_dot_general.AxisTiling(axis=2, tile_size=2)],
    )
    qx, _ = quantizer.quant(
        x, calibration_axes=[0, 2], tiling_state=tiling_state
    )
    self.assertEqual(qx.qvalue.shape, (4, 4, 2, 2))
    self.assertEqual(qx.scale[0].shape, (1, 4, 2, 1))
    self.assertEqual(qx.scale[0].dtype, jnp.float32)

    x = qx.dequant()
    self.assertEqual(x.shape, (4, 4, 4))

  def test_mid_quantization(self):
    def make_binary_dg(use_mid):
      mid_alpha: str | float = 0.5 if use_mid else config.SKIP
      bits = 1
      dg = config.config_v4(
          fwd_bits=bits,
          dlhs_bits=bits,
          drhs_bits=bits,
          fwd_mid_alpha_both=mid_alpha,
          dlhs_mid_alpha_both=mid_alpha,
          drhs_mid_alpha_both=mid_alpha,
      )
      # for exact equality
      dg.fwd.dg_quantizer.lhs.numerics.preserve_max_val = True
      dg.fwd.dg_quantizer.rhs.numerics.preserve_max_val = True
      # PO2 scales for exact equality
      dg.fwd.dg_quantizer.lhs.calibration = functools.partial(
          dg.fwd.dg_quantizer.lhs.calibration, po2_scale=True
      )
      dg.fwd.dg_quantizer.rhs.calibration = functools.partial(
          dg.fwd.dg_quantizer.rhs.calibration, po2_scale=True
      )
      return dg

    # Note that we are testing with mid_alpha = 0.5, and po2 scales.
    a = jnp.array([[1.0, 2.0, 4.0], [1.0, 4.0, 16.0]])
    b = jnp.array([[4.0, 2.0, 1.0], [16.0, 4.0, 1.0]])
    ret = jnp.array([4.0, 16.0]) * 3.0
    dimension_numbers = (((1,), (1,)), ((0,), (0,)))

    # Sanity check.
    test_eq("", jax.lax.dot_general(a, b, dimension_numbers), ret)

    # Without mid quantization all values in a, b will be rounded up
    # to 4.0 or 8.0 because of binary quantization.
    ret_no_mid = jnp.array([3 * 4.0**2, 3 * 16.0**2])
    test_eq("", make_binary_dg(False)(a, b, dimension_numbers), ret_no_mid)

    # With mid scales all values in a, b will be equal to 2.0 and
    # binary quantization will be lossless.
    test_eq("", make_binary_dg(True)(a, b, dimension_numbers), ret)


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generate HLOs for bits and pieces of the model."""
import logging
import sys

from absl import app
from absl import flags
from aqt.jax_legacy.jax import hlo_utils
from aqt.jax_legacy.jax import quant_config
from aqt.jax_legacy.jax.quantization import QuantType
from aqt.jax_legacy.jax.wmt_mlperf import models
from aqt.jax_legacy.jax.wmt_mlperf import predict
from aqt.jax_legacy.jax.wmt_mlperf import training_hparams
from aqt.jax_legacy.jax.wmt_mlperf import training_hparams_generator_lib
from aqt.jax_legacy.jax.wmt_mlperf.hparams_configs.experimental import full_model_8bit_weights_only_int8
from aqt.jax_legacy.jax.wmt_mlperf.hparams_configs.experimental import small_model_8bit_weights_and_fixed_acts
from aqt.jax_legacy.jax.wmt_mlperf.hparams_configs.experimental import small_model_bfloat16
from aqt.jax_legacy.jax.wmt_mlperf.hparams_configs.leaderboard import full_model_bfloat16
from aqt.jax_legacy.utils import hparams_utils
from flax import serialization
from flax.core import pop
import jax
import jax.numpy as jnp
from jax.tools import jax_to_ir
from ml_collections import config_dict
import tensorflow.compat.v2 as tf

flags.DEFINE_string(
    'hlo_output', '/tmp/hlo.pb', help='Destination file of the requested HLO.')

flags.DEFINE_string(
    'checkpoint', None, help='Checkpoint file saved by train.py.')

FLAGS = flags.FLAGS

EOS_TOKEN = 2  # Default Sentencepiece EOS token.


def _restore_from_checkpoint(model, checkpoint_file: str):
  with tf.io.gfile.GFile(checkpoint_file, 'rb') as fp:
    checkpoint = serialization.msgpack_restore(fp.read())
    if 'target' not in checkpoint:
      raise ValueError('Invalid checkpoint %s: no top-level "target".' %
                       checkpoint_file)
    checkpoint_model = checkpoint['target']
    checkpoint_model = jax.tree.map(jnp.array, checkpoint_model)
    return serialization.from_state_dict(model, checkpoint_model)


def encoder_from_file(config: config_dict.ConfigDict,
                      batch_size: int = 8,
                      encode_length: int = 16,
                      use_bfloat16: bool = True,
                      use_xla_optimizations: bool = True):
  """Generates HLO for just the encoder of the WMT model.

  Args:
    config: A ConfigDict instance.
    batch_size: Batch size.
    encode_length: Max length of an input sentence.
    use_bfloat16: Use bfloat16 mixed precision training instead of float32.
    use_xla_optimizations: Whether to use xla optimizations.
  """
  if FLAGS.checkpoint:
    raise app.UsageError('Checkpoints not yet supported for WMT encoder.')

  input_shape = (batch_size, encode_length)
  rng = jax.random.PRNGKey(0)
  hparams = hparams_utils.load_dataclass_from_config_dict(
      training_hparams.TrainingHParams, config)
  model_hparams = hparams.model_hparams
  model = models.Encoder(
      vocab_size=32711,
      hparams=model_hparams.encoder,
      shared_embedding=None,
      use_bfloat16=use_bfloat16,
      emb_dim=model_hparams.emb_dim,
      num_heads=model_hparams.num_heads,
      qkv_dim=model_hparams.qkv_dim,
      mlp_dim=model_hparams.mlp_dim,
      max_len=encode_length,
      train=False,
      dropout_rate=0.1,
      attention_dropout_rate=0.1,
      dynamic_context=quant_config.DynamicContext(
          update_bounds=False, collect_acts_stats=False, quantize_acts=True))
  init_state = model.init(rng, jnp.ones(input_shape, jnp.float32))

  def _fn(state, inputs):
    return model.apply(state, inputs, mutable=False)

  if not use_xla_optimizations:
    computation = (
        jax.jit(_fn)
        .lower(init_state, jnp.ones(input_shape, jnp.float32))
        .compiler_ir(dialect='hlo')
    )
    hlo_utils.output_hlo(computation, FLAGS.hlo_output)

  else:

    def _wrapped_fn(inputs):
      return _fn(init_state, inputs)

    def to_shape_str(shape_tuple):
      return 'f32[%s]' % ','.join(map(str, shape_tuple))

    hlo_module_proto_str, hlo_txt = jax_to_ir.jax_to_hlo(
        _wrapped_fn,
        [('inputs', jax_to_ir.parse_shape_str(to_shape_str(input_shape)))])
    hlo_utils.output_hlo_to_file(hlo_module_proto_str, hlo_txt,
                                 FLAGS.hlo_output)


def encoder_n_32(layers: int):
  """Generates HLO for just the encoder n-layer of the WMT model.

  Args:
    layers: The number of model layers.
  """
  if FLAGS.checkpoint:
    raise app.UsageError('Checkpoints not yet supported for WMT encoder.')

  input_shape = (1, 32)
  rng = jax.random.PRNGKey(0)
  model_hparams = (
      training_hparams_generator_lib.create_base_transformer_hparams(
          embedding_weight_prec=None,
          attention_weight_prec=None,
          mlp_weight_prec=None,
          mlp_pos_inputs_prec=None,
          mlp_pos_inputs_hyper=None,
          mlp_signed_inputs_prec=None,
          mlp_signed_inputs_hyper=None,
          attention_kqv_inputs_prec=None,
          attention_kqv_inputs_hyper=None,
          attention_out_inputs_prec=None,
          attention_out_inputs_hyper=None,
          logits_inputs_prec=None,
          logits_inputs_hyper=None,
          logits_via_embeddings=True,
          attention_act_q_inputs_prec=None,
          attention_act_q_inputs_hyper=None,
          attention_act_k_inputs_prec=None,
          attention_act_k_inputs_hyper=None,
          attention_act_probs_inputs_prec=None,
          attention_act_v_inputs_prec=None,
          attention_act_v_inputs_hyper=None,
          num_layers=layers,
          emb_dim=1024,
          num_heads=1,
          qkv_dim=1024,
          mlp_dim=4096,
          quant_type=QuantType.FAKE_QUANT,
      )
  )
  model = models.Encoder(
      vocab_size=32711,
      hparams=model_hparams.encoder,
      shared_embedding=None,
      use_bfloat16=False,
      emb_dim=model_hparams.emb_dim,
      num_heads=model_hparams.num_heads,
      qkv_dim=model_hparams.qkv_dim,
      mlp_dim=model_hparams.mlp_dim,
      max_len=32,
      train=False,
      dropout_rate=0.1,
      attention_dropout_rate=0.1,
      dynamic_context=quant_config.DynamicContext(
          update_bounds=False, collect_acts_stats=False, quantize_acts=True))
  init_state = model.init(rng, jnp.ones(input_shape, jnp.float32))

  def _fn(state, inputs):
    return model.apply(state, inputs)

  computation = (
      jax.jit(_fn)
      .lower(init_state, jnp.ones(input_shape))
      .compiler_ir(dialect='hlo')
  )
  hlo_utils.output_hlo(computation, FLAGS.hlo_output)


def encoder_full_model_opt_8_16():
  """Generates HLO for just the encoder of full transformer bfloat16."""
  encoder_from_file(config=full_model_bfloat16.get_config())


def encoder_full_model_quantized_8_16():
  """Generates HLO for the encoder of full transformer weights quantized."""
  encoder_from_file(config=full_model_8bit_weights_only_int8.get_config())


def encoder_small_model_quantized_1_32():
  """Generates HLO for just the encoder of small WMT 8-bit quantized model."""
  encoder_from_file(
      config=small_model_8bit_weights_and_fixed_acts.get_config(),
      batch_size=1,
      encode_length=32,
      use_bfloat16=False,
      use_xla_optimizations=False)


def encoder_1_32_big_model():
  """Generates HLO for just the encoder 1-layer of the WMT model."""
  encoder_n_32(1)


def encoder_2_32_big_model():
  """Generates HLO for just the encoder 2-layer of the WMT model."""
  encoder_n_32(2)


def transformer(config: config_dict.ConfigDict, batch_size: int,
                encode_length: int, decode_length: int):
  """Generates HLO for the WMT model.

  Args:
    config: A ConfigDict instance.
    batch_size: Batch size.
    encode_length: Max length of an input sentence.
    decode_length: Max length of an output sentence.
  """
  encode_shape = (batch_size, encode_length)
  decode_shape = (batch_size, decode_length)
  rng = jax.random.PRNGKey(0)

  hparams = hparams_utils.load_dataclass_from_config_dict(
      training_hparams.TrainingHParams, config)
  model_hparams = hparams.model_hparams

  transformer_kwargs = dict(
      vocab_size=33708,  # WMT en
      output_vocab_size=33708,  # WMT de
      max_len=encode_length)
  model = models.Transformer(
      **transformer_kwargs,
      hparams=model_hparams,
      train=False,
      dynamic_context=quant_config.DynamicContext(
          update_bounds=False, collect_acts_stats=False, quantize_acts=True),
      should_decode=True,
      use_bfloat16=False,
      dropout_rate=0.0,
      attention_dropout_rate=0.0)
  state = jax.jit(model.init)(rng, jnp.ones(encode_shape, jnp.float32),
                              jnp.ones(decode_shape, jnp.float32))
  if FLAGS.checkpoint:
    model = _restore_from_checkpoint(model, FLAGS.checkpoint)
  state, cache = pop(state, 'cache')  # pytype: disable=attribute-error
  state, params = pop(state, 'params')  # pytype: disable=attribute-error
  input_dummy = jnp.ones(encode_shape)
  if FLAGS.checkpoint:

    def _with_weights(inputs):
      return predict.step(
          inputs,
          params,
          cache,
          state,
          EOS_TOKEN,
          decode_length,
          transformer_kwargs=transformer_kwargs,
          hparams=model_hparams,
          dynamic_context=quant_config.DynamicContext(
              update_bounds=False, quantize_acts=True))

    computation = (
        jax.jit(_with_weights).lower(input_dummy).compiler_ir(dialect='hlo')
    )
  else:

    def _without_weights(inputs, params):
      return predict.step(
          inputs,
          params,
          cache,
          state,
          EOS_TOKEN,
          decode_length,
          transformer_kwargs=transformer_kwargs,
          hparams=model_hparams,
          dynamic_context=quant_config.DynamicContext(
              update_bounds=False, quantize_acts=True))

    computation = (
        jax.jit(_without_weights)
        .lower(input_dummy, params)
        .compiler_ir(dialect='hlo')
    )

  hlo_utils.output_hlo(computation, FLAGS.hlo_output)


def transformer_32_64_big_model():
  """Generates HLO that corresponds to a bigger version of the July demo."""
  transformer(
      config=full_model_bfloat16.get_config(),
      batch_size=1,
      encode_length=32,
      decode_length=64)


def transformer_32_64():
  """Generates HLO that corresponds to the July demo."""
  transformer(
      config=small_model_bfloat16.get_config(),
      batch_size=1,
      encode_length=32,
      decode_length=64)


# Corresponds to cl/306954847.
def transformer_16_97_147_big_model():
  """Generates HLO that mirrors the initial WMT transformer drop (cl/306954847)."""
  transformer(
      config=full_model_bfloat16.get_config(),
      batch_size=16,
      encode_length=97,
      decode_length=147)


def main(argv):
  if len(argv) != 2:
    raise app.UsageError('Too many command-line arguments.')
  variants = {
      'encoder_small_model_quantized_1_32': encoder_small_model_quantized_1_32,
      'encoder_full_model_opt_8_16': encoder_full_model_opt_8_16,
      'encoder_full_model_quantized_8_16': encoder_full_model_quantized_8_16,
      'encoder_1_32_big_model': encoder_1_32_big_model,
      'encoder_2_32_big_model': encoder_2_32_big_model,
      'transformer_32_64_big_model': transformer_32_64_big_model,
      'transformer_32_64': transformer_32_64,
      'transformer_16_97_147_big_model': transformer_16_97_147_big_model,
  }
  requested_variant = argv[1]
  if requested_variant not in variants:
    logging.error('Requested variant %s does not exist.', requested_variant)
    sys.exit(1)
  variants[requested_variant]()


if __name__ == '__main__':
  app.run(main)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Identity numerics for passing through input as-is."""

from typing import Any
from aqt.jax.v2 import stochastic_rounding
from aqt.jax.v2 import utils
from aqt.jax.v2.numerics import numerics


@utils.flax_slots_kw_only_dataclass
class NoNumerics(numerics.AqtNumerics):
  """No quantization, use a native type such as bf16."""

  # TODO(lew): This is a workaround. We should separate Stochastic Rounding.
  # noise_fn has no effect in NoNumerics.
  noise_fn: None | stochastic_rounding.NoiseFn = None
  dtype: None | Any = None

  # TODO(lew): This is a hack. We treat check isinstance(NoNumerics) and treat
  # it in a special way right now. These functions are never called
  def get_dtype(self):
    return None

  def get_quant_bound(self):
    pass

  def vjp_fwd(self, x, context):
    pass

  def vjp_bwd(self, res, grad):
    pass



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for aqt.jax.quantization."""

import itertools
from unittest import mock

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax_legacy.jax import fp_cast
from aqt.jax_legacy.jax import get_bounds
from aqt.jax_legacy.jax import primitives
from aqt.jax_legacy.jax import quant_config
from aqt.jax_legacy.jax import quantization
from aqt.jax_legacy.jax import test_utils
from aqt.jax_legacy.jax.get_bounds import GetBounds
from aqt.jax_legacy.jax.quantization import QuantOps
from aqt.jax_legacy.jax.quantization import QuantType
from aqt.jax_legacy.jax.quantization import SCALE_DTYPE
from flax import linen as nn
import jax
from jax import random
import jax.numpy as jnp
import numpy as onp

fp32 = onp.float32
test_utils.configure_jax()


class QuantOpsTest(parameterized.TestCase):

  def setUp(self):
    super(QuantOpsTest, self).setUp()
    quantization.DISABLE_EPSILON_IN_SCALE_FUN_FOR_TESTING = True

  @parameterized.named_parameters(
      dict(testcase_name='prec_2', bounds=6.0, prec=2),
      dict(testcase_name='prec_4', bounds=6.0, prec=4),
      dict(testcase_name='prec_8', bounds=6.0, prec=8),
      dict(testcase_name='2_features_prec_8', bounds=[6., 12.], prec=8),
  )
  def test_attributes_create_positive(self, bounds, prec):
    bounds = jnp.array(bounds)
    relu6 = QuantOps.create_positive(bounds=bounds, prec=prec)
    onp.testing.assert_array_equal(relu6._scale, 2**prec / bounds)
    self.assertEqual(relu6._symmetric, False)
    self.assertEqual(relu6._prec, prec)

  @parameterized.named_parameters(
      dict(testcase_name='prec_2', bounds=6.0, prec=2),
      dict(testcase_name='prec_4', bounds=6.0, prec=4),
      dict(testcase_name='prec_8', bounds=6.0, prec=8),
      dict(testcase_name='2_features_prec_8', bounds=[6., 12.], prec=8),
  )
  def test_attributes_create_symmetric(self, bounds, prec):
    bounds = jnp.array(bounds)
    act_signed = QuantOps.create_symmetric(
        bounds=bounds, prec=prec, half_shift=False)
    onp.testing.assert_array_equal(act_signed._scale,
                                   (2**(prec - 1) - 1) / bounds)
    self.assertEqual(act_signed._symmetric, True)
    self.assertEqual(act_signed._prec, prec)

  @parameterized.named_parameters(
      dict(
          testcase_name='fp8_143',
          weight_range=[2.0, 64.0],
          weight_shape=(10, 1),
          fp_quant=QuantOps.FloatQuant(
              is_scaled=True,
              fp_spec=QuantOps.FloatQuant.FloatPrec(
                  exp_min=-11,
                  exp_max=4,
                  sig_bits=3,
              ),
          ),
      ),
      dict(
          testcase_name='fp8_152',
          weight_range=[2.0, 64.0],
          weight_shape=(10, 1),
          fp_quant=QuantOps.FloatQuant(
              is_scaled=True,
              fp_spec=QuantOps.FloatQuant.FloatPrec(
                  exp_min=-23,
                  exp_max=8,
                  sig_bits=2,
              ),
          ),
      ),
  )
  def test_attributes_create_weights_op_fp(
      self,
      weight_range,
      weight_shape,
      fp_quant,
  ):
    weights = jnp.array(
        fp32(onp.random.uniform(*weight_range, size=weight_shape)))
    axis = None if weight_shape[1] == 1 else 0
    weights_quant_op = QuantOps.create_weights_ops(
        w=weights,
        weight_params=QuantOps.WeightParams(
            prec=fp_quant, axis=axis, half_shift=False))
    max_weight = onp.max(abs(weights), axis=0)
    onp.testing.assert_array_equal(
        jnp.squeeze(weights_quant_op._scale),
        jnp.exp2(-jnp.floor(jnp.log2(max_weight))))
    self.assertEqual(weights_quant_op._symmetric, True)
    self.assertIs(weights_quant_op._prec, fp_quant)
    weights_scaled = (weights * weights_quant_op._scale).astype(weights.dtype)
    weights_quant_expected = fp_cast.downcast_sat_ftz(
        weights_scaled,
        fp_quant.fp_spec.exp_min,
        fp_quant.fp_spec.exp_max,
        fp_quant.fp_spec.sig_bits,
    )
    weights_quant_calculated = weights_quant_op.to_quantized(
        weights, dtype=SCALE_DTYPE)
    onp.testing.assert_array_equal(weights_quant_expected,
                                   weights_quant_calculated)
    # Test the lower (23 - fp_quant.fp_spec.sig_bits) bits of the calculated
    # quantized weights are zero.
    sig_mask = jnp.int32((1 << (23 - fp_quant.fp_spec.sig_bits)) - 1)
    onp.testing.assert_array_equal(
        weights_quant_calculated.view(jnp.int32) & sig_mask,
        jnp.zeros_like(weights))

  @parameterized.named_parameters(
      dict(
          testcase_name='fp_act_symmetric',
          act_distribution='symmetric',
          use_hparams_bounds=False,
      ),
      # TODO(b/193561347): FP quantization with positive input distribution is
      # not supported yet
      dict(
          testcase_name='fp_act_positive',
          act_distribution='positive',
          use_hparams_bounds=False,
      ),
      dict(
          testcase_name='fp_act_symmetric_hyper_bounds',
          act_distribution='symmetric',
          use_hparams_bounds=True,
      ),
      dict(
          testcase_name='fp_act_positive_hyper_bounds',
          act_distribution='positive',
          use_hparams_bounds=True,
      ),
  )
  def test_attributes_create_acts_op_fp(
      self,
      act_distribution,
      use_hparams_bounds,
  ):
    inputs = jnp.array(fp32(2.0 * onp.random.uniform(0, 1.0, size=(10, 4))))
    fp_quant = QuantOps.FloatQuant(
        is_scaled=True,
        fp_spec=QuantOps.FloatQuant.FloatPrec(
            exp_min=-15,
            exp_max=15,
            sig_bits=2,
        ),
    )
    if use_hparams_bounds:
      bounds = get_bounds.GetBounds.Hyper(
          initial_bound=6.0,
          stddev_coeff=1,
          absdev_coeff=0,
          mix_coeff=1,
          reset_stats=True,
          ema_coeff=None,
          use_cams=False,
          granularity=quant_config.QuantGranularity.PER_TENSOR)
    else:
      bounds = 6.0

    hparams = QuantOps.ActHParams(
        input_distribution=act_distribution,
        bounds=bounds,
        prec=fp_quant,
        half_shift=False)

    class TestModule(nn.Module):
      hparams: QuantOps.ActHParams

      @nn.compact
      def __call__(self, inputs):
        return QuantOps.create_input_ops(
            inputs,
            hparams=hparams,
            bounds_params=GetBounds.Params(
                update_stats=False, update_bounds=False))

    test_module = TestModule(hparams=hparams)
    state = test_module.init(jax.random.PRNGKey(0), inputs=inputs)
    act_quant_op = test_module.apply(state, inputs=inputs)

    act_scaled = (inputs * act_quant_op._scale).astype(inputs.dtype)
    act_quant_expected = fp_cast.downcast_sat_ftz(
        act_scaled,
        fp_quant.fp_spec.exp_min,
        fp_quant.fp_spec.exp_max,
        fp_quant.fp_spec.sig_bits,
    )
    act_quant_calculated = act_quant_op.to_quantized(inputs, dtype=SCALE_DTYPE)
    onp.testing.assert_array_equal(act_quant_expected, act_quant_calculated)

  @parameterized.named_parameters(
      dict(
          testcase_name='pos_weight_prec_2',
          weight_range=[2.0, 10.0],
          weight_shape=(10, 1),
          prec=2),
      dict(
          testcase_name='pos_weight_prec_4',
          weight_range=[2.0, 10.0],
          weight_shape=(10, 1),
          prec=4),
      dict(
          testcase_name='pos_weight_prec_8',
          weight_range=[2.0, 10.0],
          weight_shape=(10, 1),
          prec=8),
      dict(
          testcase_name='neg_weight_prec_8',
          weight_range=[-12.0, 2.0],
          weight_shape=(10, 1),
          prec=8),
      dict(
          testcase_name='neg_weight_2_features_prec_8',
          weight_range=[-12.0, 2.0],
          weight_shape=(10, 2),
          prec=8),
  )
  def test_attributes_create_weights_ops(self, weight_range, weight_shape,
                                         prec):
    weights = jnp.array(
        fp32(
            onp.random.uniform(
                weight_range[0], weight_range[1], size=weight_shape)))
    axis = 0 if weight_shape[1] != 1 else None
    weights_quant = QuantOps.create_weights_ops(
        w=weights,
        weight_params=QuantOps.WeightParams(
            prec=prec, axis=axis, half_shift=False))
    max_weight = onp.max(abs(weights), axis=0)
    onp.testing.assert_array_equal(
        jnp.squeeze(weights_quant._scale), (2**(prec - 1) - 1) / max_weight)
    self.assertEqual(weights_quant._symmetric, True)
    self.assertEqual(weights_quant._prec, prec)


class WeightQuantizationTest(parameterized.TestCase):

  @parameterized.named_parameters(
      dict(testcase_name='per_layer_quant', axis=None),
      dict(testcase_name='per_channel_quant', axis=(0,)))
  def test_weight_scale_shape_is_expected(self, axis):
    # Tests if scale is as expected for weights quantization.

    num_features = 4
    expected_scale_shape = (1, 1) if axis is None else (1, num_features)

    # Weight Quantization
    weights = jnp.array(
        fp32(2.0 * onp.random.uniform(0, 1.0, size=(10, num_features))))
    _ = QuantOps.create_weights_fake_quant(
        w=weights,
        weight_params=QuantOps.WeightParams(
            prec=8.0,
            axis=axis,
            expected_scale_shape=expected_scale_shape,
            half_shift=False))

  @parameterized.named_parameters(
      dict(testcase_name='prec_4', prec=4),
      dict(testcase_name='prec_8', prec=8))
  def test_float_weights_quantization(self, prec):
    # Tests that quantized and rescaled float weights are close to original
    # weights.
    weights = jnp.array(fp32(2.0 * onp.random.uniform(0, 1.0, size=(10, 1))))
    rescaled_weights = QuantOps.create_weights_fake_quant(
        w=weights,
        weight_params=QuantOps.WeightParams(
            prec=prec, axis=None, half_shift=False))
    test_utils.assert_all_close_prec(weights, rescaled_weights, prec=prec)

  @parameterized.named_parameters(
      dict(testcase_name='prec_2', prec=2),  #
      dict(testcase_name='prec_4', prec=4),
      dict(testcase_name='prec_8', prec=8))
  def test_full_range_int_weight_quantization(self, prec):
    # Integer weights in full range [-maxmin_signed_int, maxmin_signed_int]
    # quantizes correctly.
    minval = -2**(prec - 1) + 1
    maxval = 2**(prec - 1) - 1
    weights = random.randint(random.PRNGKey(0), (10, 1), minval, maxval + 1)
    weights = weights.at[0, :].set(maxval)
    weight_quant = QuantOps.create_weights_ops(
        w=weights,
        weight_params=QuantOps.WeightParams(
            prec=prec, axis=None, half_shift=False))
    quantized_weights = weight_quant.to_quantized(weights, dtype=SCALE_DTYPE)
    onp.testing.assert_array_equal(quantized_weights[0],
                                   (2**(prec - 1.0) - 1.0))
    rescaled_weights = weight_quant.from_quantized(
        quantized_weights, dtype=jnp.float32)
    onp.testing.assert_array_equal(weights, rescaled_weights)

  @parameterized.named_parameters(
      dict(testcase_name='prec_2', prec=2),  #
      dict(testcase_name='prec_4', prec=4),
      dict(testcase_name='prec_8', prec=8))
  def test_scale_invariance_weight_quantization(self, prec):
    # Scaling weights by power of 2, should scale the output by the same scale.
    weights = random.uniform(random.PRNGKey(0), (10, 1))
    weight_scale = 16
    scaled_weights = weights * weight_scale

    weights = QuantOps.create_weights_fake_quant(
        w=weights,
        weight_params=QuantOps.WeightParams(
            prec=prec, axis=None, half_shift=False))

    scaled_weights = QuantOps.create_weights_fake_quant(
        w=scaled_weights,
        weight_params=QuantOps.WeightParams(
            prec=prec, axis=None, half_shift=False))

    onp.testing.assert_array_equal(weights * weight_scale, scaled_weights)

  @parameterized.named_parameters(
      dict(testcase_name='prec_2', prec=2),  #
      dict(testcase_name='prec_4', prec=4),
      dict(testcase_name='prec_8', prec=8))
  def test_per_feature_dim_scale_invariance_weight_quantization(self, prec):
    # Scaling each channel of weights by a different power of 2, should scale
    # the respective channel of output by the same scale.
    weights = random.uniform(random.PRNGKey(0), (3, 4))
    weight_scale = 2**jnp.arange(4)[jnp.newaxis, :]
    scaled_weights = weights * weight_scale

    weights = quantization.QuantOps.create_weights_fake_quant(
        w=weights,
        weight_params=QuantOps.WeightParams(
            prec=prec, axis=0, half_shift=False))

    scaled_weights = quantization.QuantOps.create_weights_fake_quant(
        w=scaled_weights,
        weight_params=QuantOps.WeightParams(
            prec=prec, axis=0, half_shift=False))

    onp.testing.assert_array_equal(weights * weight_scale, scaled_weights)


class ActQuantizationTest(parameterized.TestCase):

  def test_inputs_scale_shape_is_expected(self):
    # Inputs quantization
    inputs = jnp.array(fp32(2.0 * onp.random.uniform(0, 1.0, size=(10, 4))))
    bounds = 6.0
    expected_inputs_scale_shape = ()

    _ = QuantOps.create_inputs_fake_quant(
        inputs=inputs,
        hparams=QuantOps.ActHParams(
            input_distribution=QuantOps.ActHParams.InputDistribution.SYMMETRIC,
            bounds=bounds,
            prec=8.0,
            half_shift=False),
        bounds_params=GetBounds.Params(
            update_stats=False,
            update_bounds=False,
            expected_bounds_shape=expected_inputs_scale_shape))

  @parameterized.named_parameters(
      dict(testcase_name='prec_2',
           prec=2), dict(testcase_name='prec_4', prec=4),
      dict(testcase_name='prec_8', prec=8))
  def test_positive_activation_quantization_clips_outside_bounds(self, prec):
    # Activation values less than 0 get clipped to 0, and values greater than
    # upper_bound get clipped to upper_bound
    relu6 = QuantOps.create_positive(bounds=6.0, prec=prec)
    activation = jnp.array(fp32([-0.5, 6.2, 3.141]))
    quantized_activations = relu6.to_quantized(activation, dtype=SCALE_DTYPE)
    onp.testing.assert_array_equal(quantized_activations[0:2],
                                   [0.0, 2**prec - 1])
    activations = relu6.from_quantized(quantized_activations, dtype=jnp.float32)
    max_clipped_val = (2**prec - 1) * (6.0 / 2**prec)
    onp.testing.assert_array_equal(activations[0:2], [0.0, max_clipped_val])

  @parameterized.named_parameters(
      dict(testcase_name='prec_2', prec=2),  #
      dict(testcase_name='prec_4', prec=4),
      dict(testcase_name='prec_8', prec=8))
  def test_per_feature_dim_unsigned_activation_quantization_clips_outside_bounds(
      self, prec):
    # Activation values less than -upper_bound get clipped to -upper_bound, and
    # values greater than upper_bound get clipped to upper_bound
    act_quant = QuantOps.create_symmetric(
        bounds=jnp.array([[6.0, 8.0]]), prec=prec, half_shift=False)
    activation = jnp.array(fp32([[-7, -8.9], [6.2, 9.4], [0, 0.]]))
    quantized_activations = act_quant.to_quantized(
        activation, dtype=SCALE_DTYPE)
    onp.testing.assert_array_equal(
        quantized_activations,
        jnp.array([[-2**(prec - 1.0) + 1.0], [2**(prec - 1.0) - 1.0], [0.0]]) *
        jnp.array([[1., 1.]]))
    activations = act_quant.from_quantized(
        quantized_activations, dtype=jnp.float32)
    onp.testing.assert_array_equal(activations,
                                   [[-6.0, -8.0], [6.0, 8.], [0, 0.]])

  @parameterized.named_parameters(
      dict(testcase_name='prec_2', prec=2),  #
      dict(testcase_name='prec_4', prec=4),
      dict(testcase_name='prec_8', prec=8))
  def test_scale_invariance_signed_activation_quantization(self, prec):
    # Scaling activation by power of 2 and bounds by same factor,
    # should scale the output by the same scale.
    activations = random.uniform(random.PRNGKey(0), (10, 1))
    act_scale = 8.
    scaled_activations = activations * act_scale

    bounds = 6.

    activations = QuantOps.create_inputs_fake_quant(
        inputs=activations,
        bounds_params=GetBounds.Params(
            update_stats=False, update_bounds=False),
        hparams=QuantOps.ActHParams(
            input_distribution=QuantOps.ActHParams.InputDistribution.SYMMETRIC,
            bounds=bounds,
            prec=prec,
            half_shift=False))

    scaled_activations = QuantOps.create_inputs_fake_quant(
        inputs=scaled_activations,
        bounds_params=GetBounds.Params(
            update_stats=False, update_bounds=False),
        hparams=QuantOps.ActHParams(
            input_distribution=QuantOps.ActHParams.InputDistribution.SYMMETRIC,
            bounds=bounds * act_scale,
            prec=prec,
            half_shift=False))
    onp.testing.assert_array_equal(activations * act_scale, scaled_activations)

  @parameterized.named_parameters(
      dict(testcase_name='prec_2', prec=2),  #
      dict(testcase_name='prec_4', prec=4),
      dict(testcase_name='prec_8', prec=8))
  def test_per_feature_dim_scale_invariance_pos_activation_quantization(
      self, prec):
    # Scaling each channel of activations by a different power of 2 and upper
    # bound with same scale, should scale the respective channel of output by
    # the same scale.
    activations = random.uniform(random.PRNGKey(0), (3, 4))
    act_scale = 2**jnp.arange(4)
    scaled_activations = activations * act_scale[jnp.newaxis, :]

    upper_bound = 6.0 * jnp.ones((3, 4), jnp.float32)

    act_quant_ops = QuantOps.create_positive(bounds=upper_bound, prec=prec)
    activations = act_quant_ops.fake_quant(
        activations, quantized_type=SCALE_DTYPE)

    scaled_act_quant_ops = QuantOps.create_positive(
        bounds=upper_bound * act_scale[jnp.newaxis, :], prec=prec)
    scaled_activations = scaled_act_quant_ops.fake_quant(
        scaled_activations, quantized_type=SCALE_DTYPE)
    onp.testing.assert_array_equal(activations * act_scale[jnp.newaxis, :],
                                   scaled_activations)

  @parameterized.named_parameters(
      dict(testcase_name='prec_4', prec=4),
      dict(testcase_name='prec_8', prec=8))
  def test_int_positive_act_quantization(self, prec):
    # Integer activations within upper_bound and upper_bound == 2^i s.t. i<prec
    # quantizes correctly.
    upper_bound = 2**(prec - 3)
    activations = random.randint(random.PRNGKey(0), (10, 1), 0, upper_bound)

    rescaled_activations = QuantOps.create_inputs_fake_quant(
        inputs=activations,
        bounds_params=GetBounds.Params(
            update_stats=False, update_bounds=False),
        hparams=QuantOps.ActHParams(
            input_distribution=QuantOps.ActHParams.InputDistribution.POSITIVE,
            bounds=upper_bound,
            prec=prec,
            half_shift=False))
    onp.testing.assert_array_equal(activations, rescaled_activations)

  @parameterized.named_parameters(
      dict(testcase_name='prec_2', prec=2),  #
      dict(testcase_name='prec_4', prec=4),
      dict(testcase_name='prec_8', prec=8))
  def test_int_symmetric_act_quantization(self, prec):
    # Integer activations within bounds and abs(bounds) == 2^(prec -1) - 1
    # quantizes correctly.
    bounds = 2**(prec - 1) - 1
    activations = random.randint(random.PRNGKey(0), (10, 1), -bounds, bounds)
    rescaled_activations = QuantOps.create_inputs_fake_quant(
        inputs=activations,
        bounds_params=GetBounds.Params(
            update_stats=False, update_bounds=False),
        hparams=QuantOps.ActHParams(
            input_distribution=QuantOps.ActHParams.InputDistribution.SYMMETRIC,
            bounds=bounds,
            prec=prec,
            half_shift=False))

    onp.testing.assert_array_equal(activations, rescaled_activations)

  @parameterized.named_parameters(
      dict(
          testcase_name='fp_prec_scaled',
          prec=QuantOps.FloatQuant(
              is_scaled=True,
              fp_spec=QuantOps.FloatQuant.FloatPrec(
                  exp_min=-11,
                  exp_max=4,
                  sig_bits=3,
              ),
          ),
      ),
      dict(
          testcase_name='fp_prec_unscaled',
          prec=QuantOps.FloatQuant(
              is_scaled=False,
              fp_spec=QuantOps.FloatQuant.FloatPrec(
                  exp_min=-11,
                  exp_max=4,
                  sig_bits=3,
              ),
          ),
      ),
      dict(
          testcase_name='int_prec',
          prec=4.0,
      ),
  )
  def test_no_quantization(self, prec):
    # If initial_bound==-1 when using GetBounds, then create_inputs_fake_quant
    # should be a no-op.
    inputs = jnp.array([[.3, 1.4], [-5.2, 4.0]])
    bounds = get_bounds.GetBounds.Hyper(
        initial_bound=-1,
        stddev_coeff=1,
        absdev_coeff=0,
        mix_coeff=1,
        reset_stats=True,
        ema_coeff=None,
        use_cams=False,
        granularity=quant_config.QuantGranularity.PER_TENSOR)
    hparams = quantization.QuantOps.ActHParams(
        input_distribution='symmetric',
        bounds=bounds,
        prec=prec,
        half_shift=False)

    # The call to create_inputs_fake_quant has to occur from within a Flax
    # module since it calls GetBounds, which is itself a Flax module.
    # Thus we create a wrapper module for testing.
    class TestModule(nn.Module):

      hparams: quantization.QuantOps.ActHParams

      @nn.compact
      def __call__(self, inputs):
        return quantization.QuantOps.create_inputs_fake_quant(
            inputs,
            hparams=hparams,
            bounds_params=GetBounds.Params(
                update_stats=True, update_bounds=False))

    test_module = TestModule(hparams=hparams)
    state = test_module.init(jax.random.PRNGKey(0), inputs=inputs)
    inputs_after_fake_quant, _ = test_module.apply(
        state, inputs=inputs, mutable=True)
    onp.testing.assert_array_equal(inputs, inputs_after_fake_quant)


# TODO(shivaniagrawal): Add tests for auto clip activation quantizations.


class AQTTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    key1, key2 = jax.random.split(jax.random.PRNGKey(0), 2)
    self.rhs = jax.random.normal(key1, (2, 4)) * 20
    self.lhs = jax.random.normal(key2, (3, 2)) * 2 + 3

  @parameterized.named_parameters(
      dict(
          testcase_name='per_layer_act_per_column_weight',
          act_bounds=4.0,
          weight_prec=16,
          weight_axis=(0,),
      ),
      dict(
          testcase_name='per_column_act_per_column_weight',
          act_bounds=[[3.0, 4.0]],
          weight_prec=16,
          weight_axis=(0,)),
      dict(
          testcase_name='per_layer_act_per_layer_weight',
          act_bounds=4.0,
          weight_prec=16,
          weight_axis=None),
      dict(
          testcase_name='per_column_act_per_layer_weight',
          act_bounds=[[3.0, 4.0]],
          weight_prec=16,
          weight_axis=None),
      dict(
          testcase_name='per_layer_act_no_weight_quant',
          act_bounds=4.0,
          weight_prec=None,
          weight_axis=None),
      dict(
          testcase_name='per_column_act_no_weight_quant',
          act_bounds=[[3.0, 4.0]],
          weight_prec=None,
          weight_axis=None),
      dict(
          testcase_name='no_act_quant_per_column_weight',
          act_bounds=None,
          weight_prec=16,
          weight_axis=(0,)),
      dict(
          testcase_name='no_act_quant_no_weight_quant',
          act_bounds=None,
          weight_prec=None,
          weight_axis=None),
  )
  def test_quantized_dot_aqt(self, act_bounds, weight_prec, weight_axis):
    # With a high enough precision, we expect results from fakequant and AQT to
    # be very similar.
    weight_params = QuantOps.WeightParams(
        prec=weight_prec, axis=weight_axis, half_shift=False)

    if act_bounds is None:
      act_params = None
    else:
      act_params = QuantOps.ActHParams(
          input_distribution='symmetric',
          bounds=jnp.array(act_bounds),
          prec=16,
          half_shift=False)

    def quantized_matmul(quant_type):
      return quantization.quantized_dot(
          w=self.rhs,
          act=self.lhs,
          weight_params=weight_params,
          act_hparams=act_params,
          bounds_params=None,
          quant_type=quant_type,
          prefer_int8_to_int32_dot=True)

    aqt_result = quantized_matmul(QuantType.AQT)
    fakequant_result = quantized_matmul(QuantType.FAKE_QUANT)
    onp.testing.assert_allclose(
        aqt_result,
        fakequant_result,
        rtol=1e-2,
        err_msg='AQT and fakequant significantly disagree')

  @parameterized.named_parameters(
      dict(
          testcase_name='per_layer_act_per_column_weight',
          act_bounds=4.0,
          weight_prec=16,
          weight_axis=(0,),
      ),
      dict(
          testcase_name='per_column_act_per_column_weight',
          act_bounds=[[[3.0, 4.0]]],
          weight_prec=16,
          weight_axis=(0,)),
      dict(
          testcase_name='per_layer_act_per_layer_weight',
          act_bounds=4.0,
          weight_prec=16,
          weight_axis=None),
      dict(
          testcase_name='per_column_act_per_layer_weight',
          act_bounds=[[[3.0, 4.0]]],
          weight_prec=16,
          weight_axis=None),
      dict(
          testcase_name='per_layer_act_no_weight_quant',
          act_bounds=4.0,
          weight_prec=None,
          weight_axis=None),
      dict(
          testcase_name='per_column_act_no_weight_quant',
          act_bounds=[[[3.0, 4.0]]],
          weight_prec=None,
          weight_axis=None),
      dict(
          testcase_name='no_act_quant_per_column_weight',
          act_bounds=None,
          weight_prec=16,
          weight_axis=(0,)),
      dict(
          testcase_name='no_act_quant_no_weight_quant',
          act_bounds=None,
          weight_prec=None,
          weight_axis=None),
  )
  def test_quantized_dot_general_aqt(self, act_bounds, weight_prec,
                                     weight_axis):
    # With a high enough precision, we expect results from fakequant and AQT to
    # be very similar.
    weight_params = QuantOps.WeightParams(
        prec=weight_prec, axis=weight_axis, half_shift=False)

    if act_bounds is None:
      act_params = None
    else:
      act_params = QuantOps.ActHParams(
          input_distribution='symmetric',
          bounds=jnp.array(act_bounds),
          prec=16,
          half_shift=False)

    lhs_ndims_3 = jnp.array(
        fp32(2.0 * onp.random.uniform(0, 1.0, size=(4, 3, 2))))

    def quantized_matmul(quant_type):
      return quantization.quantized_dot_general(
          w=self.rhs,
          act=lhs_ndims_3,
          weight_params=weight_params,
          act_hparams=act_params,
          bounds_params=None,
          quant_type=quant_type,
          dimension_numbers=(((lhs_ndims_3.ndim - 1,), (0,)), ((), ())),
          prefer_int8_to_int32_dot=True)

    aqt_result = quantized_matmul(QuantType.AQT)
    self.assertEqual(aqt_result.shape, (4, 3, 4))

    fakequant_result = quantized_matmul(QuantType.FAKE_QUANT)
    onp.testing.assert_allclose(
        aqt_result,
        fakequant_result,
        rtol=1e-2,
        err_msg='AQT and fakequant significantly disagree')

  def assert_is_integer_in_range(self, x, *, prec, distribution):
    if distribution == 'symmetric':
      x_clipped = primitives.round_and_clip_to_signed_int(
          x, prec=prec, dtype=x.dtype, half_shift=False)
    elif distribution == 'positive':
      x_clipped = primitives.floor_and_clip_to_unsigned_int(
          x, prec=prec, dtype=x.dtype, half_shift=False)
    else:
      raise ValueError(f'Invalid distribution {distribution}')
    onp.testing.assert_array_equal(
        x, x_clipped,
        f'Array cannot be losslessly cast to integer with precision {prec} '
        f'and {distribution} distribution.')

  @parameterized.parameters(
      dict(act_distribution='symmetric', prefer_int8_to_int32_dot=True, prec=4),
      dict(act_distribution='symmetric', prefer_int8_to_int32_dot=True, prec=8),
      dict(act_distribution='positive', prefer_int8_to_int32_dot=True, prec=4),
      dict(act_distribution='positive', prefer_int8_to_int32_dot=True, prec=8),
      dict(
          act_distribution='symmetric', prefer_int8_to_int32_dot=False, prec=4))
  @mock.patch.object(jax.lax, 'dot_general', wraps=jax.lax.dot_general)
  def test_lax_dot_has_integer_inputs_in_quantized_dot(self, mock_dot_general,
                                                       act_distribution,
                                                       prefer_int8_to_int32_dot,
                                                       prec):
    weight_params = QuantOps.WeightParams(
        prec=prec, axis=(0,), half_shift=False)
    act_params = QuantOps.ActHParams(
        input_distribution=act_distribution,
        bounds=jnp.array([[3.0, 1.5]]),
        prec=prec,
        half_shift=False)
    act = self.lhs
    if act_distribution == 'positive':
      act = jnp.abs(act)
    # We need this context manager to stop Jax from trying to compile the arms
    # of the `lax.cond` call in `dot_general_aqt`. By default, Jax will always
    # try to compile the functions passed to `lax.cond`, even if outside of a
    # JITed context. JIT compilation is incompatible with using a mock for the
    # call to 'dot_general' because during compilation Jax will expect
    # 'dot_general' to return a tracer and will throw an error if it returns a
    # mock instead. By explicily using jax.disable_jit, Jax will not try to
    # compile the arms to lax.cond and so using a mock will work fine.
    with jax.disable_jit():
      quantization.quantized_dot(
          w=self.rhs,
          act=act,
          weight_params=weight_params,
          act_hparams=act_params,
          bounds_params=None,
          quant_type=QuantType.AQT,
          prefer_int8_to_int32_dot=prefer_int8_to_int32_dot)
    act_inputs, weight_inputs = mock_dot_general.call_args[0]
    self.assert_is_integer_in_range(
        act_inputs, prec=prec, distribution=act_distribution)
    self.assert_is_integer_in_range(
        weight_inputs, prec=prec, distribution='symmetric')
    if prefer_int8_to_int32_dot and not (act_distribution == 'positive' and
                                         prec == 8):
      expected_input_dtype = jnp.int8
    else:
      expected_input_dtype = jnp.float32
    self.assertEqual(act_inputs.dtype, expected_input_dtype)
    self.assertEqual(weight_inputs.dtype, expected_input_dtype)

  @parameterized.parameters(
      itertools.product(
          (jnp.bfloat16, jnp.float32), (4, None),
          (quantization.QuantType.AQT, quantization.QuantType.FAKE_QUANT)))
  def test_quantized_dot_has_correct_dtype(self, input_dtype, act_prec,
                                           quant_type):
    weight_params = QuantOps.WeightParams(prec=4, axis=(0,), half_shift=False)
    act_params = QuantOps.ActHParams(
        input_distribution='symmetric',
        bounds=jnp.array([[3.0, 1.5]]),
        prec=act_prec,
        half_shift=False)
    act = self.lhs.astype(input_dtype)
    w = self.rhs.astype(input_dtype)
    output = quantization.quantized_dot(
        w=w,
        act=act,
        weight_params=weight_params,
        act_hparams=act_params,
        bounds_params=None,
        quant_type=quant_type,
        prefer_int8_to_int32_dot=True)
    self.assertEqual(output.dtype, input_dtype)

  @parameterized.parameters(
      itertools.product(
          (jnp.bfloat16, jnp.float32), (4, None),
          (quantization.QuantType.AQT, quantization.QuantType.FAKE_QUANT)))
  def test_dynamic_quantized_dot_general_has_correct_dtype(
      self, input_dtype, act_prec, quant_type):
    lhs_params = QuantOps.ActHParams(
        input_distribution='symmetric',
        bounds=2.0,
        prec=act_prec,
        half_shift=False)
    rhs_params = QuantOps.ActHParams(
        input_distribution='symmetric',
        bounds=1.5,
        prec=act_prec,
        half_shift=False)
    lhs_act = self.lhs.astype(input_dtype)
    rhs_act = self.rhs.astype(input_dtype)
    output = quantization.quantized_dynamic_dot_general(
        lhs_act=lhs_act,
        rhs_act=rhs_act,
        lhs_act_hparams=lhs_params,
        rhs_act_hparams=rhs_params,
        lhs_bounds_params=None,
        rhs_bounds_params=None,
        dot_dimension_numbers=(((1,), (0,)), ((), ())),
        quant_type=quant_type)
    self.assertEqual(output.dtype, input_dtype)

  @parameterized.parameters(
      dict(lhs_prec=16, rhs_prec=16), dict(lhs_prec=None, rhs_prec=16),
      dict(lhs_prec=16, rhs_prec=None), dict(lhs_prec=None, rhs_prec=None))
  def test_quantized_dynamic_dot_general(self, lhs_prec, rhs_prec):
    lhs_bounds = 2.0
    rhs_bounds = 1.5
    lhs_params = QuantOps.ActHParams(
        input_distribution='symmetric',
        bounds=lhs_bounds,
        prec=lhs_prec,
        half_shift=False)
    rhs_params = QuantOps.ActHParams(
        input_distribution='symmetric',
        bounds=rhs_bounds,
        prec=rhs_prec,
        half_shift=False)

    def quantized_matmul(quant_type):
      return quantization.quantized_dynamic_dot_general(
          lhs_act=self.lhs,
          rhs_act=self.rhs,
          lhs_act_hparams=lhs_params,
          rhs_act_hparams=rhs_params,
          lhs_bounds_params=None,
          rhs_bounds_params=None,
          dot_dimension_numbers=(((1,), (0,)), ((), ())),
          quant_type=quant_type)

    aqt_result = quantized_matmul(QuantType.AQT)
    fakequant_result = quantized_matmul(QuantType.FAKE_QUANT)
    onp.testing.assert_allclose(
        aqt_result,
        fakequant_result,
        rtol=1e-2,
        err_msg='AQT and fakequant significantly disagree')

  def test_quantized_dynamic_dot_general_get_bounds(self):

    class TestModule(nn.Module):

      @nn.compact
      def __call__(self, lhs, rhs):
        lhs_get_bounds = GetBounds.Hyper(
            initial_bound=10.0,
            stddev_coeff=0,
            absdev_coeff=0,
            mix_coeff=0,
            granularity=quant_config.QuantGranularity.PER_TENSOR)
        rhs_get_bounds = GetBounds.Hyper(
            initial_bound=5.0,
            stddev_coeff=0,
            absdev_coeff=0,
            mix_coeff=0,
            granularity=quant_config.QuantGranularity.PER_TENSOR)
        lhs_params = QuantOps.ActHParams(
            input_distribution='symmetric',
            bounds=lhs_get_bounds,
            prec=8,
            half_shift=False)
        rhs_params = QuantOps.ActHParams(
            input_distribution='symmetric',
            bounds=rhs_get_bounds,
            prec=8,
            half_shift=False)
        lhs_bounds_params = get_bounds.GetBounds.Params(
            update_stats=True, update_bounds=False, module_name='lhs')
        rhs_bounds_params = get_bounds.GetBounds.Params(
            update_stats=True, update_bounds=False, module_name='rhs')
        out = quantization.quantized_dynamic_dot_general(
            lhs_act=lhs,
            rhs_act=rhs,
            lhs_act_hparams=lhs_params,
            rhs_act_hparams=rhs_params,
            dot_dimension_numbers=(((1,), (0,)), ((), ())),
            quant_type=QuantType.AQT,
            lhs_bounds_params=lhs_bounds_params,
            rhs_bounds_params=rhs_bounds_params)
        return out

    lhs = jnp.array([[2.0]])
    rhs = jnp.array([[3.0]])
    module = TestModule()
    state = module.init(jax.random.PRNGKey(0), lhs, rhs)
    out, _ = module.apply(state, lhs, rhs, mutable=True)
    lhs_scale = 127.0 / 10.0
    rhs_scale = 127.0 / 5.0
    expected_out = (round(lhs_scale * 2.0) * round(rhs_scale * 3.0)) / (
        lhs_scale * rhs_scale)
    onp.testing.assert_allclose(out, [[expected_out]])

  @parameterized.parameters(
      dict(lhs_distribution='symmetric', rhs_distribution='symmetric'),
      dict(lhs_distribution='positive', rhs_distribution='symmetric'),
      dict(lhs_distribution='symmetric', rhs_distribution='positive'),
      dict(lhs_distribution='positive', rhs_distribution='positive'))
  @mock.patch.object(jax.lax, 'dot_general')
  def test_lax_dot_has_integer_inputs_in_dynamic_dot_general(
      self, mock_dot_general, lhs_distribution, rhs_distribution):
    lhs_params = QuantOps.ActHParams(
        input_distribution=lhs_distribution,
        bounds=2.0,
        prec=4,
        half_shift=False)
    rhs_params = QuantOps.ActHParams(
        input_distribution=rhs_distribution,
        bounds=1.5,
        prec=4,
        half_shift=False)
    lhs_act = self.lhs
    if lhs_distribution == 'positive':
      lhs_act = jnp.abs(lhs_act)
    rhs_act = self.rhs
    if rhs_distribution == 'positive':
      rhs_act = jnp.abs(rhs_act)
    quantization.quantized_dynamic_dot_general(
        lhs_act=lhs_act,
        rhs_act=rhs_act,
        lhs_act_hparams=lhs_params,
        rhs_act_hparams=rhs_params,
        lhs_bounds_params=None,
        rhs_bounds_params=None,
        dot_dimension_numbers=(((1,), (0,)), ((), ())),
        quant_type=QuantType.AQT)
    lhs_inputs, rhs_inputs = mock_dot_general.call_args[0]
    self.assert_is_integer_in_range(
        lhs_inputs, prec=4, distribution=lhs_distribution)
    self.assert_is_integer_in_range(
        rhs_inputs, prec=4, distribution=rhs_distribution)

  def test_quantized_dot_no_quant(self):
    act_hparams = QuantOps.ActHParams(
        input_distribution='symmetric', bounds=-1.0, prec=4, half_shift=False)
    weight_params = QuantOps.WeightParams(prec=4, axis=(0,), half_shift=False)
    act = jnp.array([[-5.0]])
    w = jnp.array([[-4.99]])
    res = quantization.quantized_dot(
        w=w,
        act=act,
        quant_type=quantization.QuantType.AQT,
        weight_params=weight_params,
        act_hparams=act_hparams,
        bounds_params=None,
        prefer_int8_to_int32_dot=True)
    onp.testing.assert_allclose(res, act * w)

  def test_quantized_dynamic_dot_general_no_quant(self):
    act_hparams = QuantOps.ActHParams(
        input_distribution='symmetric', bounds=-1.0, prec=4, half_shift=False)
    lhs_act = jnp.array([[-5.0]])
    rhs_act = jnp.array([[-4.99]])
    res = quantization.quantized_dynamic_dot_general(
        lhs_act=lhs_act,
        rhs_act=rhs_act,
        quant_type=quantization.QuantType.AQT,
        lhs_act_hparams=act_hparams,
        rhs_act_hparams=act_hparams,
        lhs_bounds_params=None,
        rhs_bounds_params=None,
        dot_dimension_numbers=(((1,), (0,)), ((), ())))
    onp.testing.assert_allclose(res, lhs_act * rhs_act)


class QuantizedDotFakeQuantTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.act = jnp.ones((3, 7))
    self.weight = jnp.ones((7, 4))

  @parameterized.named_parameters(
      dict(testcase_name='no_quantization', weight_prec=None, act_prec=None),
      dict(testcase_name='weight_only_quant', weight_prec=8., act_prec=None),
      dict(testcase_name='act_only_quant', weight_prec=None, act_prec=4),
      dict(testcase_name='both_quantized', weight_prec=4, act_prec=8),
      dict(
          testcase_name='both_quantized_fq_int',
          weight_prec=4,
          act_prec=8,
          strategy=QuantType.FAKE_QUANT_WITH_INT),
  )
  @mock.patch.object(QuantOps, 'create_weights_fake_quant')
  @mock.patch.object(QuantOps, 'create_inputs_fake_quant')
  def test_quantized_dot_general_should_call_weights_and_inputs_quantization(
      self,
      mock_act_fq,
      mock_w_fq,
      weight_prec,
      act_prec,
      strategy=QuantType.FAKE_QUANT):
    mock_w_fq.side_effect = lambda inputs, **_: inputs
    mock_act_fq.side_effect = lambda inputs, **_: inputs

    weight_params = QuantOps.WeightParams(
        prec=weight_prec, axis=None, half_shift=False)
    act_hparams = QuantOps.ActHParams(  # pylint: disable=g-long-ternary
        bounds=6.,
        prec=act_prec,
        input_distribution=QuantOps.ActHParams.InputDistribution.SYMMETRIC,
        half_shift=False) if act_prec else None
    bounds_params = GetBounds.Params(
        update_stats=False, update_bounds=False)

    quantization.quantized_dot(
        w=self.weight,
        act=self.act,
        quant_type=strategy,
        weight_params=weight_params,
        act_hparams=act_hparams,
        bounds_params=bounds_params,
        prefer_int8_to_int32_dot=True)

    quantized_type = strategy.to_jax_type()

    mock_w_fq.assert_called_with(
        mock.ANY,
        weight_params=weight_params,
        quantized_type=quantized_type,
        fake_dependency=mock.ANY)
    if act_hparams:
      mock_act_fq.assert_called_with(
          mock.ANY, hparams=act_hparams, bounds_params=bounds_params)
    else:
      mock_act_fq.assert_not_called()


class QuantizedDynamicDotGeneralTest(parameterized.TestCase):

  def setUp(self):
    super(QuantizedDynamicDotGeneralTest, self).setUp()
    self.lhs_act = jnp.ones((4, 2, 3, 7))
    self.rhs_act = jnp.ones((3, 7, 5, 6))
    self.dimension_numbers = (((2, 3), (0, 1)), ((), ()))

  @parameterized.named_parameters(
      dict(
          testcase_name='no_quantization', lhs_act_prec=None,
          rhs_act_prec=None),
      dict(testcase_name='lhs_only_quant', lhs_act_prec=8., rhs_act_prec=None),
      dict(testcase_name='rhs_only_quant', lhs_act_prec=None, rhs_act_prec=4),
      dict(testcase_name='both_quantized', lhs_act_prec=4, rhs_act_prec=8),
      dict(
          testcase_name='both_quantized_fq_int',
          lhs_act_prec=4,
          rhs_act_prec=8,
          strategy=QuantType.FAKE_QUANT_WITH_INT),
  )
  @mock.patch.object(QuantOps, 'create_inputs_fake_quant')
  def test_quantized_dynamic_dot_general_should_call_inputs_quantization(
      self,
      mock_act_fq,
      lhs_act_prec,
      rhs_act_prec,
      strategy=QuantType.FAKE_QUANT):
    mock_act_fq.side_effect = lambda inputs, hparams, bounds_params: inputs

    # pylint: disable=g-long-ternary
    lhs_act_hparams = QuantOps.ActHParams(
        bounds=6.,
        prec=lhs_act_prec,
        input_distribution=QuantOps.ActHParams.InputDistribution.SYMMETRIC,
        half_shift=False) if lhs_act_prec else None
    rhs_act_hparams = QuantOps.ActHParams(
        bounds=6.,
        prec=rhs_act_prec,
        input_distribution=QuantOps.ActHParams.InputDistribution.SYMMETRIC,
        half_shift=False) if rhs_act_prec else None
    # pylint: enable=g-long-ternary

    bounds_params = GetBounds.Params(
        update_stats=False, update_bounds=False)

    quantization.quantized_dynamic_dot_general(
        lhs_act=self.lhs_act,
        rhs_act=self.rhs_act,
        quant_type=strategy,
        dot_dimension_numbers=self.dimension_numbers,
        lhs_act_hparams=lhs_act_hparams,
        lhs_bounds_params=bounds_params,
        rhs_act_hparams=rhs_act_hparams,
        rhs_bounds_params=bounds_params,
    )
    calls = []
    for prec in [lhs_act_prec, rhs_act_prec]:
      if prec is not None:
        act_hparams = QuantOps.ActHParams(
            bounds=6., prec=prec, input_distribution=mock.ANY, half_shift=False)
        calls.append(
            mock.call(
                mock.ANY,
                hparams=act_hparams,
                bounds_params=bounds_params))
    self.assertLen(calls, mock_act_fq.call_count)
    mock_act_fq.assert_has_calls(calls, any_order=True)


class QuantizedSumTest(parameterized.TestCase):

  @parameterized.parameters(
      # This roughly corresponds to float32, so we expect no difference vs a
      # float32 sum.
      dict(exp_min=-2**7, exp_max=2**7, sig_bits=23, expected_result=100.001),
      # In this low precision case, the addition of .001 to the accumulator will
      # have no effect after quantization
      dict(exp_min=-2**3, exp_max=2**3, sig_bits=1, expected_result=100.0))
  def test_quantized_sum(self, exp_min, exp_max, sig_bits, expected_result):
    x = jnp.array([0.001, 100.0])
    prec = QuantOps.FloatQuant.FloatPrec(exp_min, exp_max, sig_bits)
    x_quantized_sum, x_grad = jax.value_and_grad(quantization.quantized_sum)(
        x, axis=0, keepdims=False, prec=prec)
    onp.testing.assert_allclose(
        x_quantized_sum, onp.array(expected_result), rtol=1e-6)
    # This tests that the gradient is using the straight-through-estimator
    onp.testing.assert_equal(x_grad, onp.array([1.0, 1.0]))

  @parameterized.parameters(
      dict(keepdims=True, axis=(0, 1), expected_shape=(1, 1)),
      dict(keepdims=False, axis=(0, 1), expected_shape=()),
      dict(keepdims=True, axis=(0,), expected_shape=(1, 2)),
      dict(keepdims=False, axis=(1,), expected_shape=(3,)))
  def test_keepdims_and_axis(self, keepdims, axis, expected_shape):
    x = jnp.arange(6).reshape((3, 2)).astype(jnp.float32)
    prec = QuantOps.FloatQuant.FloatPrec(-2**7, 2**7, 23)
    x_quantized_sum = quantization.quantized_sum(
        x, keepdims=keepdims, axis=axis, prec=prec)
    self.assertEqual(x_quantized_sum.shape, expected_shape)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Accurate Quantized Training ops."""

from __future__ import annotations

from typing import List, Optional, Tuple

from aqt.common import aqt_common
from aqt.common import aqt_config
from flax import linen as nn
from flax import struct
import jax
import jax.numpy as jnp


def pass_through(x: jnp.ndarray, fn) -> jnp.ndarray:
  # Create an exactly-zero expression with Sterbenz lemma that has an
  # exactly-one gradient.
  return x - jax.lax.stop_gradient(x) + jax.lax.stop_gradient(fn(x))


class Stats(nn.Module):
  """Manages efficient gatherting of running statistics."""

  # This type annotation is what the constructor expects.
  # However nn.Module silently converts it into immutable Tuple, so
  # the methods will observe Tuple[Optional[int], ...]
  data_shape: List[Optional[int]] = struct.field(pytree_node=False)
  config: aqt_config.StatsConfig = struct.field(pytree_node=False)

  def setup(self):
    """Constructor to init a Stats instance.

    Returns:
      A new instance of Stats, with initialized statistics.
    """
    self.config.validate(self.data_shape)
    # TODO(jihwanlee): Move the check below in the config validator.
    if self.config.lp_order > 30:
      raise NotImplementedError('For higher norms we should add stabilization.')
    stats_shape = list(self.data_shape)  # nn.Module converts lists to tuples.
    for axis in self.config.share_stats_axes:
      stats_shape[axis] = 1
    self.stats_shape = stats_shape

    def divide_no_nan(x: jnp.ndarray, y: jnp.ndarray) -> jnp.ndarray:
      res = jnp.where(y != 0, x / y, jnp.zeros_like(x))
      return res

    self.divide = divide_no_nan if self.config.safe_divide else jnp.divide

    def init_var(init_val: float) -> jnp.ndarray:
      return jnp.full(self.stats_shape, init_val, dtype=jnp.float32)

    def mk_var(name: str, init_val: float) -> nn.Variable:
      return self.variable('aqt', name, init_var, init_val)

    self.ema_update_count = self.config.ema_update_count
    self.sum_of_ones = mk_var('sum_of_ones', self.config.update_count_prior)
    self.sum_of_vals = mk_var(
        'sum_of_vals', self.config.mean_prior * self.config.update_count_prior
    )
    self.max_of_abs_vals = mk_var('max_of_abs_vals', self.config.max_dev_prior)
    self.sum_of_l1_vals = mk_var(
        'sum_of_l1_vals',
        self.config.l1_dev_prior * self.config.update_count_prior,
    )
    self.sum_of_lp_vals = mk_var(
        'sum_of_lp_vals',
        self.config.lp_dev_prior**self.config.lp_order
        * self.config.update_count_prior,
    )

  def __call__(self):
    # Keep __call__() just not to require users to explicitly specify which
    # method should be called when invoking init().
    pass

  # TODO(lew): Remvoe override_ema_update_count.
  #   override_ema_update_count is used only in tests and is a hack.
  #   However replacing it with  self.ema_update_count does not work,
  #   because nn.Model will not allow to access it from outside.
  #   I'm not sure how this should be solved cleanly.
  def update(
      self,
      x: jnp.ndarray,
      weight: Optional[jnp.ndarray],
      override_ema_update_count=None,
  ):
    """Returns a new Stats object with updated statistics.

    Since flax.struct.dataclass objects are frozen, thie update method uses
    the dataclass.replace method to get a new object replacing the existing
    statistics with new ones and returns it.

    Args:
      x: input array to update the current statistics with.
      weight: weight for the input array x.
      override_ema_update_count: use this value instead of the one in config.

    Returns:
      Anew object of Stats with updated statistics.
    """
    aqt_common.check_shapes_conformal(x.shape, self.data_shape)
    if weight is not None and len(x.shape) != len(weight.shape):
      raise ValueError(
          f'expected rank(x)={len(x.shape)} == rank(weight)={weight.shape}'
      )

    def update_var(var, s, reduce_max=False):
      assert len(s.shape) == len(self.data_shape), (s.shape, self.data_shape)
      assert s.size > 0
      if weight is not None:
        if reduce_max:
          # A maximum is a maximum, regardless of its nonzero weight.
          s = jnp.where(weight > 0, s, 0)
        else:
          s = s * weight
      reduce_fn = jnp.max if reduce_max else jnp.sum
      s = reduce_fn(s, axis=self.config.share_stats_axes, keepdims=True)
      if self.config.tpu_cross_replica_sum:
        raise NotImplementedError(
            'support for tpu_cross_replica_sum=True is not implemented'
        )
      override = override_ema_update_count
      rate = 1.0 / (override if override is not None else self.ema_update_count)
      var.value = (1.0 - rate) * var.value + rate * s

    # Layers such as ReLU emit zeros often. In such cases, we can model
    # the non-sparse distribution of weights separately, resulting in
    # unbiased estimation of non-sparse mean l1 and lp.
    # This clips away less of the distribution of inputs.
    if self.config.filter_zeros:
      ones = jnp.where(x != 0, 1.0, 0.0)
    else:
      ones = jnp.ones_like(x)

    px = x if self.config.lp_order % 2 == 0 else jnp.abs(x)

    update_var(self.sum_of_ones, ones)
    update_var(self.sum_of_vals, x)
    update_var(self.max_of_abs_vals, jnp.abs(x), reduce_max=True)
    update_var(self.sum_of_l1_vals, jnp.abs(x))
    update_var(self.sum_of_lp_vals, px**self.config.lp_order)

  def mean(self) -> jnp.ndarray:
    return self.divide(self.sum_of_vals.value, self.sum_of_ones.value)

  def max_dev(self) -> jnp.ndarray:
    return self.max_of_abs_vals.value

  def l1_dev(self) -> jnp.ndarray:
    return self.divide(self.sum_of_l1_vals.value, self.sum_of_ones.value)

  def lp_dev(self) -> jnp.ndarray:
    if self.config.lp_order == 2:
      # sqrt() is numerically more accurate
      return jnp.sqrt(
          self.divide(self.sum_of_lp_vals.value, self.sum_of_ones.value)
      )
    else:
      # TODO(b/205769820): Make sure if the output of pow op below is
      # numerically valid.
      return self.divide(self.sum_of_lp_vals.value, self.sum_of_ones.value) ** (
          1.0 / self.config.lp_order
      )

  def bound(  #
      self, calibration_config: aqt_config.CalibrationConfig
  ) -> jnp.ndarray:
    """Compute the upper bound on input tensor values, broadcastable to input."""
    return (
        calibration_config.l1_dev_coeff * self.l1_dev()
        + calibration_config.lp_dev_coeff * self.lp_dev()  #
        + calibration_config.max_dev_coeff * self.max_dev()  #
        + calibration_config.const_bound_coeff  #
    )


def is_config_active(
    config: aqt_config.AqtTensorConfig, event_count: jnp.ndarray
) -> bool:
  """Return whether the configuration is active at event_count."""
  config.validate()
  should_q = True
  if config.begin_at_event is not None:
    should_q &= config.begin_at_event <= event_count
  if config.end_at_event is not None:
    should_q &= event_count < config.end_at_event
  return should_q  # pytype: disable=bad-return-type  # jnp-type


class TensorQuantizer(nn.Module):
  """Maintains state associated with the quantization of an input tensor.

  A TensorQuantizer owns observed statistics for an input tensor, along with
  variables for the scale and event_count copy, recording the last time of then
  most recent update to this `TensorQuantizer` class.

  This class provides state-mutating methods for updating statistics for every
  observation of an input tensor and is used by AQT methods to derive
  calibration bounds.

  TensorQuantizer assumes the supplied `event_count` strictly monotonically
  increases across `TensorQuantizer.update` calls and starts out strictly
  greater than `int32.min`.

  Attributes:
    data_shape: the shape of input tensor. Some dimensions may be of unknown
      size, but these must have stats shared in `config.stats_config`. In this
      case, use of quanitzed variable is disallowed.
    config: A training quantization schedule.
    quantized_variable: If provided during initialization, stores the quantized
      version of the observed tensor from the most recent `update()`.
  """

  data_shape: List[Optional[int]]
  config: Optional[aqt_config.AqtScheduleConfig] = None

  def setup(self):
    if self.config is not None:
      self.config.fill_gaps_with_float_config()
      self.config.validate(self.data_shape)
      self._stats = Stats(self.data_shape, self.config.stats_config)
      self._scale = self.variable(
          'TensorQuantizer',
          'scale',
          jnp.zeros,
          self._stats.stats_shape,
          jnp.float32,
      )
      self._inv_scale = self.variable(
          'TensorQuantizer',
          'inv_scale',
          jnp.zeros,
          self._stats.stats_shape,
          jnp.float32,
      )
      if self.config.use_quantized_variable:
        self.quantized_variable = self.variable(
            'TensorQuantizer',
            'quantized_variable',
            jnp.zeros,
            self.data_shape,
            jnp.int8,
        )
      else:
        self.quantized_variable = self.variable(
            'TensorQuantizer', 'quantized_variable', lambda: None
        )
      self._last_update = self.variable(
          'TensorQuantizer', 'last_update', lambda: jnp.iinfo(jnp.int32).min
      )

  def __call__(self):
    # Keep __call__() just not to require users to explicitly specify which
    # method should be called when invoking init().
    pass

  def _fresh_scale(
      self, config: aqt_config.AqtTensorConfig
  ) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Returns an updated scale provided by a config or the current one if None."""
    if isinstance(config.quant_config, aqt_config.FloatConfig):
      # We shouldn't update the scale if the given config contains FloatConfig;
      # fill with a poison value if we get into this situation.
      nan = jnp.full(self._stats.stats_shape, jnp.nan, dtype=jnp.float32)
      return nan, nan

    x_bound = self._stats.bound(config.calibration_config)
    clip_bound = aqt_common.get_clip_bound(config.quant_config)

    new_scale = clip_bound / x_bound
    inv_scale = x_bound / clip_bound
    return new_scale, inv_scale

  def clip_range(self) -> jnp.ndarray:
    """Returns the tensor clip range or zeros if no int config is active."""
    if self.config is None:
      raise ValueError('config is None')
    for config in self.config.tensor_configs:
      if is_config_active(config, self._last_update.value) and isinstance(
          config.quant_config, aqt_config.IntQuantConfig
      ):
        clip_bound = aqt_common.get_clip_bound(config.quant_config)
        return self._inv_scale.value * clip_bound
    return jnp.zeros(self._stats.stats_shape)

  def update(
      self,  #
      sample: jnp.ndarray,
      weight: Optional[jnp.ndarray],
      event_count: jnp.ndarray,
  ):
    """Updates statistics, scale, and quantized variable."""

    if self.config is None:
      return

    active_configs = [
        c
        for c in self.config.tensor_configs
        if is_config_active(c, event_count)
    ]

    if len(active_configs) == 1:
      self._update_config(active_configs[0], sample, weight, event_count)
    else:
      raise ValueError('There must be exactly one active config.')

  def _update_config(
      self,  # ,
      config: aqt_config.AqtTensorConfig,
      sample: jnp.ndarray,
      weight: Optional[jnp.ndarray],
      event_count: jnp.ndarray,
  ):
    """update(), but with active `config`."""

    should_update_scale = self._should_update_scale(config, event_count)
    self._last_update.value = event_count

    self._stats.update(sample, weight)

    new_scale, inv_scale = jax.lax.cond(
        should_update_scale,  #
        lambda: self._fresh_scale(config),
        lambda: (self._scale.value, self._inv_scale.value),
    )
    self._scale.value = new_scale
    self._inv_scale.value = inv_scale
    if self.config is None:
      raise ValueError('config is None')
    if (
        self.config.use_quantized_variable
        and isinstance(config.quant_config, aqt_config.IntQuantConfig)
        and config.quant_config.compatible_with_int8()
    ):
      scale, _ = self._get_quant_scale(train=True)
      sample = scale * sample
      new_var = self._to_quant(sample, train=True).astype(
          self.quantized_variable.value.dtype
      )
      self.quantized_variable.value = new_var

  def _should_update_scale(
      self, config: aqt_config.AqtTensorConfig, event_count: jnp.ndarray  #
  ) -> bool:
    """Returns if scale should be updated for the config and event count."""
    if isinstance(config.quant_config, aqt_config.FloatConfig):
      return False

    if not config.freeze_scale_at_begin:
      return True

    # The first time a config is active, even if we freeze scale, we should
    # update the scale.
    was_previously_inactive = not is_config_active(
        config, self._last_update.value
    )

    # We rely on jnp.int32.min being an illegal event count value, so that
    # even if is_config_active(config, jnp.int32.min), we still update scale.
    # This could happen if the very first update happens for a config which
    # has freeze_scale_at_begin and begin=None.
    assert (
        event_count > jnp.iinfo(jnp.int32).min
    ), 'event_count cannot be int32.min'

    first_event = jnp.array(self._last_update.value == jnp.iinfo(jnp.int32).min)

    return was_previously_inactive | first_event  # pytype: disable=bad-return-type  # jnp-type

  def _to_quant(self, x: jnp.ndarray, train: bool) -> jnp.ndarray:
    """Quantizes x with active quant config, if any, else act as identity."""

    def qparams(
        config: aqt_config.AqtTensorConfig, check_active: bool = True
    ) -> Tuple[bool, jnp.ndarray, jnp.ndarray, jnp.ndarray]:
      """Returns parameters for AQT quantization routine."""

      should_quantize = False
      clip_bound = jnp.array(0.0)
      shift_before = jnp.array(0.0)
      shift_after = jnp.array(0.0)

      config_active = is_config_active(config, self._last_update.value)
      config_active = not check_active or config_active

      if isinstance(config.quant_config, aqt_config.FloatConfig):
        clip_bound = jnp.where(config_active, float('inf'), 0.0)
        return should_quantize, clip_bound, shift_before, shift_after
      elif isinstance(config.quant_config, aqt_config.IntQuantConfig):
        should_quantize |= config_active
        config_active = jnp.float32(config_active)

        clip_bound += config_active * aqt_common.safe_clip_bound(
            config.quant_config
        )

        if config.quant_config.preserve_zero:
          shift_before += config_active * 0.5
        else:
          shift_after += config_active * 0.5

        return should_quantize, clip_bound, shift_before, shift_after
      else:
        raise NotImplementedError(
            'Only supporting FloatConfig and IntQuantConfig for now.'
        )

    if self.config is None:
      clip_bound = jnp.array(float('inf'))
      shift_before = shift_after = 0.0
      should_quantize = False
    elif not train and self.config.inference_config_index is not None:
      should_quantize, clip_bound, shift_before, shift_after = qparams(
          self.config.tensor_configs[self.config.inference_config_index],
          check_active=False,
      )
    elif self.config.tensor_configs:
      should_quantize, clip_bound, shift_before, shift_after = zip(
          *map(qparams, self.config.tensor_configs)
      )
      should_quantize = any(should_quantize)
      clip_bound = sum(clip_bound)
      shift_before = sum(shift_before)
      shift_after = sum(shift_after)
    else:
      clip_bound = shift_before = shift_after = 0.0
      should_quantize = False

    maybe_floor = lambda y: jnp.where(should_quantize, jnp.floor(y), y)

    # Note that backprop does not depend directly on the value of _last_update
    # or any_config_active; only scales and constants need to be maintained
    # and there's no branching on ops including the input tensor x. This
    # results in significant memory reduction, see cl/415355150.
    x = jnp.clip(x, -clip_bound, clip_bound)
    x += shift_before
    x = pass_through(x, maybe_floor)
    # TODO(b/219778053): Add a test that validates the shift of values. Mutants
    # found that removing the line below doesn't affect any tests.
    x += shift_after

    return x

  def _get_quant_scale(self, train: bool) -> jnp.ndarray:
    """Returns scales to quantize/dequantize the active quant config, if any, else ones."""
    if self.config is None:
      scale = inv_scale = jnp.array(1.0, dtype=jnp.float32)
      return scale, inv_scale  # pytype: disable=bad-return-type  # jax-ndarray

    if not train and self.config.inference_config_index is not None:
      inference_config = self.config.tensor_configs[
          self.config.inference_config_index
      ]
      should_quantize = isinstance(
          inference_config.quant_config, aqt_config.IntQuantConfig
      )
    else:
      should_quantize = False
      for config in self.config.tensor_configs:
        if isinstance(config.quant_config, aqt_config.FloatConfig):
          continue
        config_active = is_config_active(config, self._last_update.value)
        should_quantize |= config_active

    scale = jnp.where(
        should_quantize, self._scale.value, jnp.ones_like(self._scale.value)
    )
    inv_scale = jnp.where(
        should_quantize,
        self._inv_scale.value,
        jnp.ones_like(self._inv_scale.value),
    )

    return scale, inv_scale  # pytype: disable=bad-return-type  # jax-ndarray
